__package__ = 'txrestapi'

from six import PY2, b

import re
import json
import base64
import os.path
import doctest

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.defer import inlineCallbacks
from twisted.web.resource import Resource, NoResource
from twisted.web.server import Request, Site
from twisted.web.client import getPage
from twisted.trial import unittest

from .resource import APIResource
from .json_resource import JsonAPIResource
from .methods import GET, PUT


class FakeChannel(object):
    transport = None

    def getPeer(self):
        return None

    def getHost(self):
        return None


def getRequest(method, url):
    req = Request(FakeChannel(), None)
    req.method = method
    req.path = url
    return req


class APIResourceTest(unittest.TestCase):

    test_class = APIResource

    def test_returns_normal_resources(self):
        r = self.test_class()
        a = Resource()
        r.putChild(b('a'), a)
        req = Request(FakeChannel(), None)
        a_ = r.getChild(b('a'), req)
        self.assertEqual(a, a_)

    def test_registry(self):
        compiled = re.compile(b('regex'))
        r = self.test_class()
        r.register(b('GET'), b('regex'), None)
        self.assertEqual([x[0] for x in r._registry], [b('GET')])
        self.assertEqual(r._registry[0][0], b('GET'))
        # This doesn't work:
        #     FailTest: <_sre.SRE_Pattern object at 0x1052f6990> != <_sre.SRE_Pattern object at 0x1052f6d78>
        # self.assertEqual(r._registry[0][1], compiled)
        self.assertEqual(r._registry[0][1].pattern, compiled.pattern)
        self.assertEqual(r._registry[0][2], None)

    def test_method_matching(self):
        r = self.test_class()
        r.register(b('GET'), b('regex'), 1)
        r.register(b('PUT'), b('regex'), 2)
        r.register(b('GET'), b('another'), 3)

        req = getRequest(b('GET'), b('regex'))
        result = r._get_callback(req)
        self.assert_(result)
        self.assertEqual(result[0], 1)

        req = getRequest(b('PUT'), b('regex'))
        result = r._get_callback(req)
        self.assert_(result)
        self.assertEqual(result[0], 2)

        req = getRequest(b('GET'), b('another'))
        result = r._get_callback(req)
        self.assert_(result)
        self.assertEqual(result[0], 3)

        req = getRequest(b('PUT'), b('another'))
        result = r._get_callback(req)
        self.assertEqual(result, (None, None))

    def test_callback(self):
        marker = object()

        def cb(request):
            return marker

        r = self.test_class()
        r.register(b('GET'), b('regex'), cb)
        req = getRequest(b('GET'), b('regex'))
        result = r.getChild(b('regex'), req)
        self.assertEqual(result.render(req), marker)

    def test_longerpath(self):
        marker = object()
        r = self.test_class()

        def cb(request):
            return marker

        r.register(b('GET'), b('/regex/a/b/c'), cb)
        req = getRequest(b('GET'), b('/regex/a/b/c'))
        result = r.getChild(b('regex'), req)
        self.assertEqual(result.render(req), marker)

    def test_args(self):
        r = self.test_class()

        def cb(request, **kwargs):
            return kwargs

        r.register(b('GET'), b('/(?P<a>[^/]*)/a/(?P<b>[^/]*)/c'), cb)
        req = getRequest(b('GET'), b('/regex/a/b/c'))
        result = r.getChild(b('regex'), req)
        self.assertEqual(sorted(result.render(req).keys()), ['a', 'b'])

    def test_order(self):
        r = self.test_class()

        def cb1(request, **kwargs):
            kwargs.update({'cb1': True})
            return kwargs

        def cb(request, **kwargs):
            return kwargs

        # Register two regexes that will match
        r.register(b('GET'), b('/(?P<a>[^/]*)/a/(?P<b>[^/]*)/c'), cb1)
        r.register(b('GET'), b('/(?P<a>[^/]*)/a/(?P<b>[^/]*)'), cb)
        req = getRequest(b('GET'), b('/regex/a/b/c'))
        result = r.getChild(b('regex'), req)
        # Make sure the first one got it
        self.assert_('cb1' in result.render(req))

    def test_no_resource(self):
        r = self.test_class()
        r.register(b('GET'), b('^/(?P<a>[^/]*)/a/(?P<b>[^/]*)$'), None)
        req = getRequest(b('GET'), b('/definitely/not/a/match'))
        result = r.getChild(b('regex'), req)
        self.assert_(isinstance(result, NoResource))

    def test_all(self):
        r = self.test_class()

        def get_cb(r):
            return b('GET')

        def put_cb(r):
            return b('PUT')

        def all_cb(r):
            return b('ALL')

        r.register(b('GET'), b('^path'), get_cb)
        r.register(b('ALL'), b('^path'), all_cb)
        r.register(b('PUT'), b('^path'), put_cb)
        # Test that the ALL registration picks it up before the PUT one
        for method in (b('GET'), b('PUT'), b('ALL')):
            req = getRequest(method, b('path'))
            result = r.getChild(b('path'), req)
            self.assertEqual(result.render(req), b('ALL') if method == b('PUT') else method)


class JSONAPIResourceTest(APIResourceTest):

    test_class = JsonAPIResource

    def test_all(self):
        r = self.test_class()

        def get_cb(r):
            return {
                'method': b('GET'),
            }

        def put_cb(r):
            return {
                'method': b('PUT'),
            }

        def all_cb(r):
            return {
                'method': b('ALL'),
            }

        r.register(b('GET'), b('^path'), get_cb)
        r.register(b('ALL'), b('^path'), all_cb)
        r.register(b('PUT'), b('^path'), put_cb)
        # Test that the ALL registration picks it up before the PUT one
        for method in (b('GET'), b('PUT'), b('ALL')):
            req = getRequest(method, b('path'))
            result = r.getChild(b('path'), req)
            self.assertEqual(json.loads(result.render(req))['method'], b('ALL') if method == b('PUT') else method)

    def test_args(self):
        r = self.test_class()

        def cb(request, **kwargs):
            return {
                'kwargs': kwargs,
            }

        r.register(b('GET'), b('/(?P<a>[^/]*)/a/(?P<b>[^/]*)/c'), cb)
        req = getRequest(b('GET'), b('/regex/a/b/c'))
        result = r.getChild(b('regex'), req)
        self.assertEqual(sorted(json.loads(result.render(req))['kwargs'].keys()), ['a', 'b'])

    def test_callback(self):
        marker = base64.b64encode(os.urandom(20))

        def cb(request):
            return {
                'marker': marker,
            }

        r = self.test_class()
        r.register(b('GET'), b('regex'), cb)
        req = getRequest(b('GET'), b('regex'))
        result = r.getChild(b('regex'), req)
        self.assertEqual(json.loads(result.render(req))['marker'], marker)

    def test_longerpath(self):
        marker = base64.b64encode(os.urandom(20))
        r = self.test_class()

        def cb(request):
            return {
                'marker': marker,
            }

        r.register(b('GET'), b('/regex/a/b/c'), cb)
        req = getRequest(b('GET'), b('/regex/a/b/c'))
        result = r.getChild(b('regex'), req)
        self.assertEqual(json.loads(result.render(req))['marker'], marker)

    def test_no_resource(self):
        r = self.test_class()
        r.register(b('GET'), b('^/(?P<a>[^/]*)/a/(?P<b>[^/]*)$'), None)
        req = getRequest(b('GET'), b('/definitely/not/a/match'))
        result = r.getChild(b('regex'), req)
        self.assertEqual(
            json.loads(result.render(req))['errors'],
            [
                "path 'regex' not found",
            ],
        )


class TestResource(Resource):
    isLeaf = True

    def render(self, request):
        return b('aresource')


class TestAPI(APIResource):
    @GET(b('^/(?P<a>test[^/]*)/?'))
    def _on_test_get(self, request, a):
        return b('GET %s') % a

    @PUT(b('^/(?P<a>test[^/]*)/?'))
    def _on_test_put(self, request, a):
        return b('PUT %s') % a

    @GET(b('^/gettest'))
    def _on_gettest(self, request):
        return TestResource()


class DecoratorsTest(unittest.TestCase):
    def _listen(self, site):
        return reactor.listenTCP(0, site, interface='127.0.0.1')  # @UndefinedVariable

    def setUp(self):
        r = TestAPI()
        site = Site(r, timeout=None)
        self.port = self._listen(site)
        self.portno = self.port.getHost().port

    def tearDown(self):
        return self.port.stopListening()

    def getURL(self, path):
        return b('http://127.0.0.1:%d/%s' % (self.portno, path))

    @inlineCallbacks
    def test_get(self):
        url = self.getURL('test_thing/')
        result = yield getPage(url, method=b('GET'))
        self.assertEqual(result, b('GET test_thing'))

    @inlineCallbacks
    def test_put(self):
        url = self.getURL('test_thing/')
        result = yield getPage(url, method=b('PUT'))
        self.assertEqual(result, b('PUT test_thing'))

    @inlineCallbacks
    def test_resource_wrapper(self):
        url = self.getURL('gettest')
        result = yield getPage(url, method=b('GET'))
        self.assertEqual(result, b('aresource'))


def test_suite():
    import unittest as ut
    suite = unittest.TestSuite()
    suite.addTest(ut.makeSuite(DecoratorsTest))
    suite.addTest(ut.makeSuite(APIResourceTest))
    if PY2:
        suite.addTest(doctest.DocFileSuite(os.path.join('..', 'README.rst')))
    return suite



# operation: token:issue
# openfield: token_name:total_number

# operation: token:transfer
# openfield: token_name:amount

import log
from hashlib import blake2b

__version__ = '0.0.2'


def blake2bhash_generate(data):
    # new hash
    blake2bhash = blake2b(str(data).encode(), digest_size=20).hexdigest()
    return blake2bhash
    # new hash


def tokens_update(node, db_handler_instance):

    db_handler_instance.index_cursor.execute('CREATE TABLE IF NOT EXISTS tokens (block_height INTEGER, timestamp, token, address, recipient, txid, amount INTEGER)')
    db_handler_instance.index.commit()

    db_handler_instance.index_cursor.execute('SELECT block_height FROM tokens ORDER BY block_height DESC LIMIT 1;')
    try:
        token_last_block = int(db_handler_instance.index_cursor.fetchone()[0])
    except:
        token_last_block = 0

    node.logger.app_log.warning('Token anchor block: {}'.format(token_last_block))

    # node.logger.app_log.warning all token issuances
    db_handler_instance.c.execute(
        'SELECT block_height, timestamp, address, recipient, signature, operation, openfield FROM transactions WHERE block_height >= ? AND operation = ? AND reward = 0 ORDER BY block_height ASC;', (
            token_last_block,
            'token:issue',
        )
    )
    results = db_handler_instance.c.fetchall()
    node.logger.app_log.warning(results)

    tokens_processed = []

    for x in results:
        try:
            token_name = x[6].split(':')[0].lower().strip()
            try:
                db_handler_instance.index_cursor.execute('SELECT * from tokens WHERE token = ?', (token_name, ))
                dummy = db_handler_instance.index_cursor.fetchall()[0]  # check for uniqueness
                node.logger.app_log.warning('Token issuance already processed: {}'.format(token_name))
            except:
                if token_name not in tokens_processed:
                    block_height = x[0]
                    node.logger.app_log.warning('Block height {}'.format(block_height))

                    timestamp = x[1]
                    node.logger.app_log.warning('Timestamp {}'.format(timestamp))

                    tokens_processed.append(token_name)
                    node.logger.app_log.warning('Token: {}'.format(token_name))

                    issued_by = x[3]
                    node.logger.app_log.warning('Issued by: {}'.format(issued_by))

                    txid = x[4][:56]
                    node.logger.app_log.warning('Txid: {}'.format(txid))

                    total = x[6].split(':')[1]
                    node.logger.app_log.warning('Total amount: {}'.format(total))

                    db_handler_instance.index_cursor.execute('INSERT INTO tokens VALUES (?,?,?,?,?,?,?)', (block_height, timestamp, token_name, 'issued', issued_by, txid, total))

                    if node.plugin_manager:
                        node.plugin_manager.execute_action_hook('token_issue', {'token': token_name, 'issuer': issued_by, 'txid': txid, 'total': total})

                else:
                    node.logger.app_log.warning('This token is already registered: {}'.format(x[1]))
        except:
            node.logger.app_log.warning('Error parsing')

    db_handler_instance.index.commit()
    # node.logger.app_log.warning all token issuances

    # node.logger.app_log.warning("---")

    # node.logger.app_log.warning all transfers of a given token
    # token = "worthless"

    db_handler_instance.c.execute(
        'SELECT operation, openfield FROM transactions WHERE (block_height >= ? OR block_height <= ?) AND operation = ? and reward = 0 ORDER BY block_height ASC;',
        (
            token_last_block,
            -token_last_block,
            'token:transfer',
        ),
    )  #includes mirror blocks
    openfield_transfers = db_handler_instance.c.fetchall()
    # print(openfield_transfers)

    tokens_transferred = []
    for transfer in openfield_transfers:
        token_name = transfer[1].split(':')[0].lower().strip()
        if token_name not in tokens_transferred:
            tokens_transferred.append(token_name)

    if tokens_transferred:
        node.logger.app_log.warning('Token transferred: {}'.format(tokens_transferred))

    for token in tokens_transferred:
        try:
            node.logger.app_log.warning('processing {}'.format(token))
            db_handler_instance.c.execute(
                'SELECT block_height, timestamp, address, recipient, signature, operation, openfield FROM transactions WHERE (block_height >= ? OR block_height <= ?) AND operation = ? AND openfield LIKE ? AND reward = 0 ORDER BY block_height ASC;',
                (
                    token_last_block,
                    -token_last_block,
                    'token:transfer',
                    token + ':%',
                )
            )
            results2 = db_handler_instance.c.fetchall()
            node.logger.app_log.warning(results2)

            for r in results2:
                block_height = r[0]
                node.logger.app_log.warning('Block height {}'.format(block_height))

                timestamp = r[1]
                node.logger.app_log.warning('Timestamp {}'.format(timestamp))

                token = r[6].split(':')[0]
                node.logger.app_log.warning('Token {} operation'.format(token))

                sender = r[2]
                node.logger.app_log.warning('Transfer from {}'.format(sender))

                recipient = r[3]
                node.logger.app_log.warning('Transfer to {}'.format(recipient))

                txid = r[4][:56]
                if txid == '0':
                    txid = blake2bhash_generate(r)
                node.logger.app_log.warning('Txid: {}'.format(txid))

                try:
                    transfer_amount = int(r[6].split(':')[1])
                except:
                    transfer_amount = 0

                node.logger.app_log.warning('Transfer amount {}'.format(transfer_amount))

                # calculate balances
                db_handler_instance.index_cursor.execute('SELECT sum(amount) FROM tokens WHERE recipient = ? AND block_height < ? AND token = ?', (
                    sender,
                    block_height,
                    token,
                ))

                try:
                    credit_sender = int(db_handler_instance.index_cursor.fetchone()[0])
                except:
                    credit_sender = 0
                node.logger.app_log.warning("Sender's credit {}".format(credit_sender))

                db_handler_instance.index_cursor.execute('SELECT sum(amount) FROM tokens WHERE address = ? AND block_height <= ? AND token = ?', (
                    sender,
                    block_height,
                    token,
                ))
                try:
                    debit_sender = int(db_handler_instance.index_cursor.fetchone()[0])
                except:
                    debit_sender = 0
                node.logger.app_log.warning("Sender's debit: {}".format(debit_sender))
                # calculate balances

                # node.logger.app_log.warning all token transfers
                balance_sender = credit_sender - debit_sender

                node.logger.app_log.warning("Sender's balance {}".format(balance_sender))

                try:
                    db_handler_instance.index_cursor.execute('SELECT txid from tokens WHERE txid = ?', (txid, ))
                    dummy = db_handler_instance.index_cursor.fetchone()  # check for uniqueness
                    if dummy:
                        node.logger.app_log.warning('Token operation already processed: {} {}'.format(token, txid))
                    else:
                        if (balance_sender - transfer_amount >= 0 and transfer_amount > 0):
                            db_handler_instance.index_cursor.execute('INSERT INTO tokens VALUES (?,?,?,?,?,?,?)', (abs(block_height), timestamp, token, sender, recipient, txid, transfer_amount))
                            if node.plugin_manager:
                                node.plugin_manager.execute_action_hook('token_transfer', {'token': token, 'from': sender, 'to': recipient, 'txid': txid, 'amount': transfer_amount})

                        else:  # save block height and txid so that we do not have to process the invalid transactions again
                            node.logger.app_log.warning('Invalid transaction by {}'.format(sender))
                            db_handler_instance.index_cursor.execute('INSERT INTO tokens VALUES (?,?,?,?,?,?,?)', (block_height, '', '', '', '', txid, ''))
                except Exception as e:
                    node.logger.app_log.warning('Exception {}'.format(e))

                node.logger.app_log.warning('Processing of {} finished'.format(token))
        except:
            node.logger.app_log.warning('Error parsing')

        db_handler_instance.index.commit()


if __name__ == '__main__':
    from libs import node, logger
    import dbhandler

    node = node.Node()
    node.debug_level = 'WARNING'
    node.terminal_output = True

    node.logger = logger.Logger()
    node.logger.app_log = log.log('local_test.log', node.debug_level, node.terminal_output)
    node.logger.app_log.warning('Configuration settings loaded')

    db_handler = dbhandler.DbHandler('static/index_local_test.db', 'static/ledger.db', 'static/hyper.db', False, None, node.logger, False)

    tokens_update(node, db_handler)
    # tokens_update("tokens.db","reindex")



#!/usr/bin/env python
# cmd_line_json_templates.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (cmd_line_json_templates.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com

#------------------------------------------------------------------------------


def ls(tpl, tag='result'):
    return '{.section %s}{.repeated section @}%s{.end}{.end}' % (tag, tpl)


#------------------------------------------------------------------------------

tpl_status = '{.section status}{@}{.end}'
tpl_result = ls('{@}\n')
tpl_message = '{.section message}{@}{.end}'
tpl_errors = ls('{@}\n', tag='errors')
tpl_execution = '{.section execution}: {@} sec.{.end}'
tpl_4_items = """{0}{1}
{2}{3}
"""
tpl_5_items = """{0}{1}
{2}{3}{4}
"""

#------------------------------------------------------------------------------

TPL_JSON = '{@}'

#------------------------------------------------------------------------------

TPL_RAW = tpl_5_items.format(tpl_status, tpl_execution, tpl_result, tpl_message, tpl_errors)

#------------------------------------------------------------------------------

TPL_KEYS_LIST = tpl_4_items.format(tpl_status, tpl_execution, ls('{key_id}\n'), tpl_errors)

TPL_KEY_GET = tpl_4_items.format(tpl_status, tpl_execution, ls('{key_id} : {type}, {size} bits, {fingerprint}\n\n{public}\n\n{private}\n'), tpl_errors)

TPL_KEY_CREATE = tpl_5_items.format(tpl_status, tpl_execution, tpl_message, ls('\n\n[{key_id}]\ntype:{type} size:{size} fingerprint:{fingerprint}\n{public}\n{private}\n'), tpl_errors)

#------------------------------------------------------------------------------

TPL_BACKUPS_LIST = tpl_4_items.format(tpl_status, tpl_execution, ls('{global_id} {type} {size} {path} {.section versions}{.repeated section @}[{backup_id}: {size}]{.end}{.end} {.section key_id}<{@}>{.end}\n'), tpl_errors)

TPL_BACKUPS_LIST_IDS = tpl_4_items.format(tpl_status, tpl_execution, ls('{backup_id} {size} {path}\n'), tpl_errors)

TPL_BACKUPS_RUNNING_LIST = tpl_5_items.format(
    tpl_status,
    tpl_execution,
    '{.section result}\n%s\n\n%s{.end}' % (
        ls('{backup_id} from {source_path} of total {total_size}, currently {bytes_processed} bytes processed, ready by {progress}%\n', tag='running'),
        ls('{id}: {path_id} from {local_path} created at {created}\n', tag='pending'),
    ),
    tpl_message,
    tpl_errors,
)

TPL_BACKUPS_TASKS_LIST = tpl_5_items.format(tpl_status, tpl_execution, ls('{id}: {path_id} from {local_path} created at {created}\n'), tpl_message, tpl_errors)

#------------------------------------------------------------------------------

TPL_RESTORES_RUNNING_LIST = tpl_5_items.format(tpl_status, tpl_execution, ls('{backup_id}, currently {bytes_processed} bytes processed\n'), tpl_message, tpl_errors)

#------------------------------------------------------------------------------

TPL_OPTIONS_LIST_KEY_TYPE_VALUE = tpl_4_items.format(tpl_status, tpl_execution, ls('{key} ({type}) : {value}\n'), tpl_errors)

#------------------------------------------------------------------------------

TPL_OPTION_MODIFIED = tpl_4_items.format(tpl_status, tpl_execution, ls('{key} ({type}) : {value}{.section old_value}, previous : {@}{.or}{.end}\n'), tpl_errors)

TPL_OPTION_SINGLE = tpl_4_items.format(tpl_status, tpl_execution, ls('{key} ({type}) : {value}{.section old_value}, previous : {@}{.or}{.end}\n'), tpl_errors)

#------------------------------------------------------------------------------

TPL_SUPPLIERS = tpl_4_items.format(tpl_status, tpl_execution, ls('{position}: {idurl}, since {connected}, keeps {.section fragments}{files}{.end} files, {contact_state}\n'), tpl_errors)

#------------------------------------------------------------------------------

TPL_CUSTOMERS = tpl_4_items.format(tpl_status, tpl_execution, ls('{position}: {idurl}, {status}\n'), tpl_errors)

#------------------------------------------------------------------------------

TPL_STORAGE = tpl_4_items.format(
    tpl_status,
    tpl_execution,
    ls(
        """{.section consumed}consumed:
    suppliers: {suppliers_num}
    needed: {needed} ({needed_str})
    used: {used} ({used_str}) {used_percent}
    available: {available} ({available_str})
    donated from each supplier: {needed_per_supplier} ({needed_per_supplier_str})
    used on each supplier: {used_per_supplier} ({used_per_supplier_str})
    available on every supplier: {available_per_supplier} ({available_per_supplier_str})
{.end}{.section donated}donated:
    customers: {customers_num}
    donated: {donated} ({donated_str})
    allocated: {consumed} ({consumed_str}) {consumed_percent}
    uploaded: {used} ({used_str}) {used_percent}
{.section customers}{.repeated section @}      {idurl} uploaded {real_str} from {consumed_str}\n{.end}{.end}{.end}{.section local}local:
    disk size: {disktotal} ({disktotal_str})
    disk free: {diskfree} ({diskfree_str}) {diskfree_percent}
    disk consumed: {total} ({total_str}) {total_percent}
      buffered files: {backups} ({backups_str})
      temporary files: {temp} ({temp_str})
      customers files: {customers} ({customers_str})
{.end}""",
    ),
    tpl_errors,
)

#------------------------------------------------------------------------------

TPL_AUTOMATS = tpl_4_items.format(tpl_status, tpl_execution, ls('{index}: {id}({state}) {.section timers}   timers: {@}{.end}\n'), tpl_errors)

#------------------------------------------------------------------------------

TPL_SERVICES = tpl_4_items.format(tpl_status, tpl_execution, ls('{index}: {name}({state}) {enabled_label}, with {num_depends} dependencies\n'), tpl_errors)

TPL_SERVICES_TREE = tpl_4_items.format(tpl_status, tpl_execution, ls('{offset}{name}({state}) {enabled_label} [{depends}]\n'), tpl_errors)

TPL_SERVICE_INFO = tpl_4_items.format(
    tpl_status,
    tpl_execution,
    ls("""ID: {index}
name: {name}
state: {state}
enabled: {enabled}
dependent on: {.section depends}{.repeated section @}{@}{.alternates with}, {.end}{.end}
config path: {config_path}
"""),
    tpl_errors,
)

#------------------------------------------------------------------------------

TPL_FRIEND_LOOKUP = """{.section result}
friend lookup:
    {result}
    {nickname}
    {position}
    {idurl}
{.end}"""

TPL_FRIEND_LOOKUP_REPEATED_SECTION = """{.section result}
friends:
{.repeated section @}
    {idurl} : {nickname}
{.end}
{.end}"""

TPL_MESSAGE_SENDING = """{.section result}
    {@}
{.end}
{.section error}
    error: {@}
{.end}
    recipient: {.section recipient}{@}{.end}
"""



#!/usr/bin/python
# read.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (read.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#

#
# This is to read some possibly incomplete raid data and recover original data.
#
# Basic idea is to find a parity that is only missing one data, then fix that
#  data, and look at parities again to see if another is now missing only one.
#  This can be called an "iterative algorithm" or "belief propagation".
#  There are algorithms that can recover data when this can not, but
#    they are more complicated.
# Input is a file and a RAID spec.
# The spec says for each datablock which parity blocks XOR it.
# As we read in each datablock we XOR all the parity blocks for it.
# Probably best to read in all the datablocks at once, and do all
# the parities one time.  If we did one datablock after another
# the parity blocks would need to be active all the time.
#
# If we XOR integers, we have a byte order issue probably, though
# maybe not since they all stay the same order, whatever that is.
#
# Some machines are 64-bit.  Would be fun to make use of that if
# we have it.
#
# 2^1 => 3       ^ is XOR operator  - so maybe we can do it in Python
#
# In the production code, when there are multiple errors we really want
# to fix the one we can fix fastest first.  The only danger is that
# we get too many errors at the same time.  We reduce this danger if
# we fix the one we can fix fastest first.

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
import six
from io import open
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

import os
import sys

#------------------------------------------------------------------------------

if __name__ == '__main__':
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..', '..')))

#------------------------------------------------------------------------------

import bitdust.logs.lg

import bitdust.raid.eccmap

#------------------------------------------------------------------------------


def ReadBinaryFile(filename):
    if not os.path.isfile(filename):
        return ''
    if not os.access(filename, os.R_OK):
        return ''
    f = open(filename, 'rb')
    data = f.read()
    f.close()
    return data


def RebuildOne(inlist, listlen, outfilename, threshold_control=None):
    readsize = 1  # vary from 1 byte to 4 bytes
    raidfiles = ['']*listlen  # just need a list of this size
    raidreads = ['']*listlen
    for filenum in range(listlen):
        try:
            raidfiles[filenum] = open(inlist[filenum], 'rb')
        except:
            bitdust.logs.lg.exc()
            for f in raidfiles:
                try:
                    f.close()
                except:
                    pass
            return False

    rebuildfile = open(outfilename, 'wb')
    progress = 0
    while True:
        for k in range(listlen):
            raidreads[k] = raidfiles[k].read(2048)
        if not raidreads[0]:
            break
        i = 0
        while i < len(raidreads[0]):
            xor = 0
            for j in range(listlen):
                b1 = ord(raidreads[j][i:i + 1])
                xor = xor ^ b1
            if six.PY3:
                out_byte = bytes([
                    xor,
                ])
            else:
                out_byte = chr(xor)
            rebuildfile.write(out_byte)
            i += readsize
            progress += 1

            if threshold_control:
                if not threshold_control(readsize):
                    raise Exception('task cancelled')

    for filenum in range(listlen):
        raidfiles[filenum].close()
    rebuildfile.close()

    if _Debug:
        with open('/tmp/raid.log', 'a') as logfile:
            logfile.write(u'raidread.RebuildOne inlist=%d listlen=%d outfilename=%r progress=%d\n' % (len(inlist), listlen, outfilename, progress))
    return True


# If segment is good, there is a file for it, if not then no file exists.
# We only rebuild data segments.
# Could only make parity segments from existing data segments, so no help toward getting data.
# As long as we could rebuild one more data segment in a pass,
#  we do another pass to see if we are then able to do another.
#  When we can't do anything more we could how many good data segments there
#    are, and if we have all we win, if not we fail.


def raidread(
    OutputFileName,
    eccmapname,
    version,
    blockNumber,
    data_parity_dir,
    threshold_control=None,
):
    try:
        if _Debug:
            open('/tmp/raid.log', 'a').write(u'raidread OutputFileName=%s blockNumber=%s eccmapname=%s\n' % (repr(OutputFileName), blockNumber, eccmapname))

        myeccmap = bitdust.raid.eccmap.eccmap(eccmapname)
        GoodFiles = [
            '',
        ]*(myeccmap.datasegments + myeccmap.paritysegments)
        MakingProgress = 1
        while MakingProgress == 1:
            MakingProgress = 0
            for PSegNum in range(myeccmap.paritysegments):
                PFileName = os.path.join(
                    data_parity_dir,
                    version,
                    str(blockNumber) + '-' + str(PSegNum) + '-Parity',
                )
                if os.path.exists(PFileName):
                    Map = myeccmap.ParityToData[PSegNum]
                    TotalDSegs = 0
                    GoodDSegs = 0
                    BadName = ''
                    for DSegNum in Map:
                        TotalDSegs += 1
                        FileName = os.path.join(
                            data_parity_dir,
                            version,
                            str(blockNumber) + '-' + str(DSegNum) + '-Data',
                        )
                        if os.path.exists(FileName):
                            GoodFiles[GoodDSegs] = FileName
                            GoodDSegs += 1
                        else:
                            BadName = FileName
                    if GoodDSegs == TotalDSegs - 1:
                        MakingProgress = 1
                        GoodFiles[GoodDSegs] = PFileName
                        GoodDSegs += 1
                        RebuildOne(GoodFiles, GoodDSegs, BadName, threshold_control=threshold_control)

        GoodFiles = []
        #  Count up the good segments and combine
        GoodDSegs = 0
        output = open(OutputFileName, 'wb')
        for DSegNum in range(myeccmap.datasegments):
            FileName = os.path.join(
                data_parity_dir,
                version,
                str(blockNumber) + '-' + str(DSegNum) + '-Data',
            )
            if os.path.exists(FileName):
                GoodDSegs += 1
                fin = open(FileName, 'rb')
                moredata = fin.read()
                fin.close()
                output.write(moredata)
        output.close()

        if _Debug:
            open('/tmp/raid.log', 'a').write(u'GoodDSegs=%d\n' % GoodDSegs)
        return GoodDSegs

    except:
        bitdust.logs.lg.exc()
        return None


def main():
    if (len(sys.argv) < 3):
        print('raidread needs an output filename and eccmap name')
        sys.exit(2)

    OutputFileName = sys.argv[1]
    eccmapname = sys.argv[2]
    version = sys.argv[3]
    block_number = sys.argv[4]
    data_parity_dir = sys.argv[5]

    print('raidread is starting with')
    print(OutputFileName)
    print(eccmapname)
    raidread(OutputFileName, eccmapname, version, block_number, data_parity_dir)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# rsa_key.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (rsa_key.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: rsa_key.


"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10
_CryptoLog = None

#------------------------------------------------------------------------------

try:
    from Cryptodome.PublicKey import RSA
    from Cryptodome.Signature import pkcs1_15
    from Cryptodome.Cipher import PKCS1_OAEP
except:
    from Crypto.PublicKey import RSA  # @UnresolvedImport @Reimport
    from Crypto.Signature import pkcs1_15  # @UnresolvedImport @Reimport
    from Crypto.Cipher import PKCS1_OAEP  # @UnresolvedImport @Reimport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng

from bitdust.system import local_fs

from bitdust.crypt import hashes
from bitdust.crypt import number

#------------------------------------------------------------------------------


class RSAKey(object):
    def __init__(self):
        self.keyObject = None
        self.local_key_id = None
        self.label = ''
        self.signed = None
        self.active = True

    def isReady(self):
        return self.keyObject is not None

    def forget(self):
        self.keyObject = None
        self.local_key_id = None
        self.label = ''
        self.signed = None
        self.active = False
        # gc.collect()
        return True

    def size(self):
        return self.keyObject.size_in_bits()

    def generate(self, bits):
        if self.keyObject:
            raise ValueError('key object already exist')
        self.keyObject = RSA.generate(bits)
        return True

    def isPublic(self):
        if not self.keyObject:
            raise ValueError('key object is not exist')
        if self.keyObject.has_private():
            return False
        return True

    def isPrivate(self):
        if not self.keyObject:
            raise ValueError('key object is not exist')
        if not self.keyObject.has_private():
            return False
        return True

    def public(self):
        if self.isPublic():
            return self
        return self.keyObject.publickey()

    def fromDict(self, key_dict):
        if self.keyObject:
            raise ValueError('key object already exist')
        key_src = key_dict['body']
        result = self.fromString(key_src)
        if result:
            self.label = key_dict.get('label', '')
            self.active = key_dict.get('active', True)
            self.local_key_id = key_dict.get('local_key_id', None)
            if 'signature' in key_dict and 'signature_pubkey' in key_dict:
                self.signed = (
                    key_dict['signature'],
                    key_dict['signature_pubkey'],
                )
        del key_src
        # gc.collect()
        return result

    def fromString(self, key_src):
        if self.keyObject:
            raise ValueError('key object already exist')
        key_src = strng.to_bin(key_src)
        try:
            self.keyObject = RSA.import_key(key_src)  # @UndefinedVariable
        except:
            if _Debug:
                lg.exc('key_src=%r' % key_src)
            raise ValueError('failed to read key body')
        del key_src
        # gc.collect()
        return True

    def fromFile(self, keyfilename):
        if self.keyObject:
            raise ValueError('key object already exist')
        key_src = local_fs.ReadTextFile(keyfilename)
        key_src = strng.to_bin(key_src)
        try:
            self.keyObject = RSA.import_key(key_src)  # @UndefinedVariable
        except:
            if _Debug:
                lg.exc('key_src=%r' % key_src)
        del key_src
        # gc.collect()
        return True

    def toPrivateString(self, output_format='PEM'):
        if not self.keyObject:
            raise ValueError('key object is not exist')
        if not self.keyObject.has_private():
            raise ValueError('this key contains only public component')
        return strng.to_text(self.keyObject.exportKey(format=output_format))

    def toPublicString(self, output_format='OpenSSH'):
        if not self.keyObject:
            raise ValueError('key object is not exist')
        return strng.to_text(self.keyObject.publickey().exportKey(format=output_format))

    def toDict(self, include_private=False, output_format_private='PEM', output_format_public='OpenSSH'):
        if not self.keyObject:
            raise ValueError('key object is not exist')
        if include_private and not self.keyObject.has_private():
            raise ValueError('this key contains only public component')
        if include_private:
            key_body = strng.to_text(self.keyObject.exportKey(format=output_format_private))
        else:
            key_body = strng.to_text(self.keyObject.publickey().exportKey(format=output_format_public))
        key_dict = {
            'body': key_body,
            'local_key_id': self.local_key_id,
            'label': self.label,
            'active': self.active,
        }
        if self.isSigned():
            key_dict.update({
                'signature': self.signed[0],
                'signature_pubkey': self.signed[1],
            })
        return key_dict

    def sign(self, message, as_digits=True):
        global _CryptoLog
        # if _CryptoLog is None:
        #     _CryptoLog = os.environ.get('CRYPTO_LOG') == '1'
        if not self.keyObject:
            raise ValueError('key object is not exist')
        if not strng.is_bin(message):
            raise ValueError('message must be byte string')
        h = hashes.sha1(message, return_object=True)
        signature_raw = pkcs1_15.new(self.keyObject).sign(h)
        if not as_digits:
            if _Debug:
                if _CryptoLog:
                    lg.args(_DebugLevel, signature_raw=signature_raw)
            return signature_raw
        signature_long = number.bytes_to_long(signature_raw)
        signature_bytes = strng.to_bin(signature_long)
        if _Debug:
            if _CryptoLog:
                lg.args(_DebugLevel, signature_bytes=signature_bytes)
        return signature_bytes

    def verify(self, signature, message, signature_as_digits=True):
        global _CryptoLog
        # if _CryptoLog is None:
        #     _CryptoLog = os.environ.get('CRYPTO_LOG') == '1'
        signature_bytes = signature
        if signature_as_digits:
            signature_bytes = number.long_to_bytes(signature, blocksize=4)
        if not strng.is_bin(signature_bytes):
            raise ValueError('signature must be byte string')
        if not strng.is_bin(message):
            raise ValueError('message must be byte string')
        h = hashes.sha1(message, return_object=True)
        result = False
        try:
            pkcs1_15.new(self.keyObject).verify(h, signature_bytes)
            result = True
        except (
            ValueError,
            TypeError,
        ):
            # do not raise any exception... just return False
            lg.exc('signature=%r message=%r' % (signature, message))
        if _Debug:
            if _CryptoLog:
                lg.args(_DebugLevel, result=result, signature=signature)
        return result

    def encrypt(self, private_message):
        if not self.keyObject:
            raise ValueError('key object is not exist')
        cipher = PKCS1_OAEP.new(self.keyObject)
        ciphertext = cipher.encrypt(private_message)
        return ciphertext

    def decrypt(self, encrypted_payload):
        if not self.keyObject:
            raise ValueError('key object is not exist')
        cipher = PKCS1_OAEP.new(self.keyObject)
        private_message = cipher.decrypt(encrypted_payload)
        return private_message

    def isSigned(self):
        return self.signed is not None



#!/usr/bin/python
# service_employer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_employer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_employer
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return EmployerService()


class EmployerService(LocalService):

    service_name = 'service_employer'
    config_path = 'services/employer/enabled'

    def dependent_on(self):
        return [
            'service_customer',
            'service_nodes_lookup',
        ]

    def start(self):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.main.config import conf
        from bitdust.main import events
        from bitdust.raid import eccmap
        from bitdust.services import driver
        from bitdust.customer import fire_hire
        self.starting_deferred = Deferred()
        self.starting_deferred.addErrback(lambda err: lg.warn('service %r was not started: %r' % (self.service_name, err.getErrorMessage() if err else 'unknown reason')))
        self.all_suppliers_hired_event_sent = False
        if driver.is_on('service_entangled_dht'):
            self._do_join_suppliers_dht_layer()
        eccmap.Update()
        fire_hire.A('init')
        fire_hire.A().addStateChangedCallback(self._on_fire_hire_ready, None, 'READY')
        conf().addConfigNotifier('services/customer/suppliers-number', self._on_suppliers_number_modified)
        conf().addConfigNotifier('services/customer/needed-space', self._on_needed_space_modified)
        events.add_subscriber(self._on_supplier_modified, 'supplier-modified')
        events.add_subscriber(self._on_dht_layer_connected, 'dht-layer-connected')
        if fire_hire.IsAllHired():
            self.starting_deferred.callback(True)
            self.starting_deferred = None
            lg.info('all my suppliers are already hired')
            return True
        fire_hire.A('restart')
        return self.starting_deferred

    def stop(self):
        from bitdust.main.config import conf
        from bitdust.main import events
        from bitdust.customer import fire_hire
        fire_hire.A().removeStateChangedCallback(self._on_fire_hire_ready)
        events.remove_subscriber(self._on_dht_layer_connected, 'dht-layer-connected')
        events.remove_subscriber(self._on_supplier_modified, 'supplier-modified')
        conf().removeConfigNotifier('services/customer/suppliers-number')
        conf().removeConfigNotifier('services/customer/needed-space')
        fire_hire.Destroy()
        return True

    def health_check(self):
        from bitdust.raid import eccmap
        from bitdust.contacts import contactsdb
        from bitdust.userid import id_url
        missed_suppliers = id_url.empty_count(contactsdb.suppliers())
        # to have that service running minimum amount of suppliers must be already in the family
        return missed_suppliers <= eccmap.Current().CorrectableErrors

    def _do_cleanup_dht_suppliers(self, use_cache=True):
        from bitdust.logs import lg
        from bitdust.services import driver
        if driver.is_on('service_entangled_dht'):
            from bitdust.dht import dht_relations
            from bitdust.userid import my_id
            d = dht_relations.read_customer_suppliers(my_id.getIDURL(), use_cache=use_cache)
            d.addCallback(self._on_my_dht_relations_discovered)
            d.addErrback(self._on_my_dht_relations_failed)
        else:
            lg.warn('service service_entangled_dht is OFF')

    def _do_check_all_hired(self):
        from bitdust.logs import lg
        from bitdust.main import events
        from bitdust.customer import fire_hire
        if fire_hire.IsAllHired():
            self._do_cleanup_dht_suppliers()
            if not self.all_suppliers_hired_event_sent:
                lg.info('at the moment all my suppliers are hired and known')
                self.all_suppliers_hired_event_sent = True
                events.send('my-suppliers-all-hired', data=dict())
            if self.starting_deferred and not self.starting_deferred.called:
                self.starting_deferred.callback(True)
                self.starting_deferred = None
        else:
            self.all_suppliers_hired_event_sent = False
            lg.info('some of my suppliers are not hired yet, or needs to be removed')
            events.send('my-suppliers-yet-not-hired', data=dict())
            if self.starting_deferred and not self.starting_deferred.called:
                self.starting_deferred.errback(Exception('not possible to hire enough suppliers'))
                self.starting_deferred = None

    def _do_notify_supplier_position(self, supplier_idurl, supplier_position):
        from bitdust.p2p import p2p_service
        from bitdust.raid import eccmap
        from bitdust.userid import my_id
        p2p_service.SendContacts(
            remote_idurl=supplier_idurl,
            json_payload={
                'space': 'family_member',
                'type': 'supplier_position',
                'customer_idurl': my_id.getIDURL(),
                'customer_ecc_map': eccmap.Current().name,
                'supplier_idurl': supplier_idurl,
                'supplier_position': supplier_position,
            },
        )

    def _do_join_suppliers_dht_layer(self):
        from bitdust.logs import lg
        from bitdust.dht import dht_service
        from bitdust.dht import dht_records
        from bitdust.dht import known_nodes
        lg.info('going to join suppliers DHT layer: %d' % dht_records.LAYER_SUPPLIERS)
        known_seeds = known_nodes.nodes()
        dht_service.open_layer(
            seed_nodes=known_seeds,
            layer_id=dht_records.LAYER_SUPPLIERS,
            connect_now=True,
            attach=False,
        )

    def _on_fire_hire_ready(self, oldstate, newstate, evt, *args, **kwargs):
        self._do_check_all_hired()
        return None

    def _on_suppliers_number_modified(self, path, value, oldvalue, result):
        from bitdust.logs import lg
        from bitdust.customer import fire_hire
        from bitdust.raid import eccmap
        lg.info('my desired suppliers number changed')
        eccmap.Update()
        self._do_check_all_hired()
        fire_hire.ClearLastFireTime()
        fire_hire.A('restart')

    def _on_needed_space_modified(self, path, value, oldvalue, result):
        from bitdust.logs import lg
        from bitdust.customer import fire_hire
        lg.info('my needed space value modified')
        self._do_check_all_hired()
        fire_hire.ClearLastFireTime()
        fire_hire.A('restart')

    def _on_supplier_modified(self, evt):
        self._do_cleanup_dht_suppliers(use_cache=False)

    def _on_my_dht_relations_discovered(self, dht_result):
        from bitdust.p2p import p2p_service
        from bitdust.contacts import contactsdb
        from bitdust.userid import my_id
        from bitdust.userid import id_url
        from bitdust.crypt import my_keys
        from bitdust.logs import lg
        if not (dht_result and isinstance(dht_result, dict) and len(dht_result.get('suppliers', [])) > 0):
            lg.warn('no dht records found for my customer family')
            return
        if id_url.is_some_empty(contactsdb.suppliers()):
            lg.warn('some of my suppliers are not hired yet, skip doing any changes')
            return
        suppliers_to_be_dismissed = set()
        dht_suppliers = id_url.to_bin_list(dht_result['suppliers'])
        # clean up old suppliers
        for idurl in dht_suppliers:
            if not idurl:
                continue
            if not contactsdb.is_supplier(idurl):
                lg.warn('dht relation with %r is not valid anymore' % idurl)
                suppliers_to_be_dismissed.add(idurl)
        for supplier_idurl in suppliers_to_be_dismissed:
            service_info = {}
            my_customer_key_id = my_id.getGlobalID(key_alias='customer')
            if my_keys.is_key_registered(my_customer_key_id):
                service_info['customer_public_key'] = my_keys.get_key_info(
                    key_id=my_customer_key_id,
                    include_private=False,
                    include_signature=False,
                    generate_signature=False,
                )
            p2p_service.SendCancelService(
                remote_idurl=supplier_idurl,
                service_name='service_supplier',
                json_payload=service_info,
            )
        if suppliers_to_be_dismissed:
            lg.info('found %d suppliers to be cleaned and sent CancelService() packets' % len(suppliers_to_be_dismissed))

    def _on_my_dht_relations_failed(self, err):
        from bitdust.logs import lg
        lg.err(err)

    def _on_dht_layer_connected(self, evt):
        if evt.data['layer_id'] == 0:
            self._do_join_suppliers_dht_layer()



#!/usr/bin/python
# network_service.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (network_service.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: network_service.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.defer import Deferred  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.services import driver

from bitdust.interface import api

from bitdust.userid import my_id

#------------------------------------------------------------------------------


def do_service_test(service_name, result_defer, wait_timeout):
    try:
        svc_info = api.service_info(service_name)
        if not svc_info or 'result' not in svc_info:
            lg.err('failed to fetch service info: %r' % svc_info)
            result_defer.callback(dict(
                error='disconnected',
                reason='{}_info_error'.format(service_name),
            ))
            return None
        svc_state = svc_info['result']['state']
    except:
        lg.exc('service "%s" test failed' % service_name)
        result_defer.callback(dict(
            error='disconnected',
            reason='{}_info_error'.format(service_name),
        ))
        return None
    if _Debug:
        lg.args(_DebugLevel, service=service_name, state=svc_state, wait_timeout=wait_timeout)
    if svc_state == 'STARTING':
        reactor.callLater(0.1, do_service_test, service_name, result_defer, wait_timeout)  # @UndefinedVariable
        return None
    if svc_state != 'ON':
        do_service_restart(service_name, result_defer, wait_timeout)
        return None
    if service_name == 'service_network':
        reactor.callLater(0, do_service_test, 'service_gateway', result_defer, wait_timeout)  # @UndefinedVariable
    elif service_name == 'service_gateway':
        reactor.callLater(0, do_service_test, 'service_p2p_hookups', result_defer, wait_timeout)  # @UndefinedVariable
    elif service_name == 'service_p2p_hookups':
        reactor.callLater(0, do_p2p_connector_test, result_defer)  # @UndefinedVariable
    elif service_name == 'service_proxy_transport':
        reactor.callLater(0, do_service_proxy_transport_test, result_defer)  # @UndefinedVariable
    else:
        raise Exception('unknown service to test %s' % service_name)
    return None


def do_service_restart(service_name, result_defer, wait_timeout):
    if _Debug:
        lg.args(_DebugLevel, service_name=service_name)
    d = api.service_restart(service_name, wait_timeout=wait_timeout)
    d.addCallback(on_service_restarted, service_name, result_defer, wait_timeout)
    d.addErrback(lambda err: result_defer.callback(dict(
        error=err,
        reason='{}_restart_error'.format(service_name),
    ), ))
    return None


def do_service_proxy_transport_test(result_defer):
    if _Debug:
        lg.dbg(_DebugLevel, 'checking proxy_transport')
    if not driver.is_enabled('service_proxy_transport'):
        result_defer.callback({
            'service_network': 'started',
            'service_gateway': 'started',
            'service_p2p_hookups': 'started',
            'service_proxy_transport': 'disabled',
        })
        return None
    try:
        proxy_receiver_lookup = automat.find('proxy_receiver')
        if not proxy_receiver_lookup:
            lg.warn('disconnected, reason is "proxy_receiver_not_found"')
            result_defer.callback(dict(
                error='disconnected',
                reason='proxy_receiver_not_found',
            ))
            return None
        proxy_receiver_machine = automat.by_index(proxy_receiver_lookup[0])
        if not proxy_receiver_machine:
            lg.warn('disconnected, reason is "proxy_receiver_not_exist"')
            result_defer.callback(dict(
                error='disconnected',
                reason='proxy_receiver_not_exist',
            ))
            return None
        if proxy_receiver_machine.state != 'LISTEN':
            lg.warn('disconnected, reason is "proxy_receiver_disconnected", sending "start" event to proxy_receiver()')
            proxy_receiver_machine.automat('start')
            result_defer.callback(dict(
                error='disconnected',
                reason='proxy_receiver_disconnected',
            ))
            return None
        result_defer.callback({
            'service_network': 'started',
            'service_gateway': 'started',
            'service_p2p_hookups': 'started',
            'service_proxy_transport': 'started',
            'proxy_receiver_state': proxy_receiver_machine.state,
        })
    except:
        lg.exc()
        result_defer.callback(dict(
            error='disconnected',
            reason='proxy_receiver_error',
        ))
    return None


def do_p2p_connector_test(result_defer):
    if _Debug:
        lg.dbg(_DebugLevel, 'checking p2p_connector')
    try:
        p2p_connector_lookup = automat.find('p2p_connector')
        if not p2p_connector_lookup:
            lg.warn('disconnected, reason is "p2p_connector_not_found"')
            result_defer.callback(dict(
                error='disconnected',
                reason='p2p_connector_not_found',
            ))
            return None
        p2p_connector_machine = automat.by_index(p2p_connector_lookup[0])
        if not p2p_connector_machine:
            lg.warn('disconnected, reason is "p2p_connector_not_exist"')
            result_defer.callback(dict(
                error='disconnected',
                reason='p2p_connector_not_exist',
            ))
            return None
        if p2p_connector_machine.state in [
            'DISCONNECTED',
        ]:
            lg.warn('disconnected, reason is "p2p_connector_disconnected", sending "check-synchronize" event to p2p_connector()')
            p2p_connector_machine.automat('check-synchronize')
            result_defer.callback(dict(
                error='disconnected',
                reason='p2p_connector_disconnected',
            ))
            return None
        do_service_proxy_transport_test(result_defer)
    except:
        lg.exc()
        result_defer.callback(dict(
            error='disconnected',
            reason='p2p_connector_error',
        ))
    return None


#------------------------------------------------------------------------------


def on_service_restarted(resp, service_name, result_defer, wait_timeout):
    if _Debug:
        lg.args(_DebugLevel, resp=resp, service_name=service_name)
    if service_name == 'service_network':
        do_service_test('service_gateway', result_defer, wait_timeout)
    elif service_name == 'service_gateway':
        do_service_test('service_p2p_hookups', result_defer, wait_timeout)
    else:
        do_p2p_connector_test(result_defer)
    return resp


def on_service_proxy_transport_check_healthy(healthy, wait_timeout):
    if _Debug:
        lg.args(_DebugLevel, healthy=healthy)
    if healthy is True:
        return None
    lg.err('service_proxy_transport is not healthy, going to restart it now')
    d = Deferred()
    d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='network_service.on_service_proxy_transport_check_healthy', ignore=True)
    do_service_restart('service_proxy_transport', d, wait_timeout=wait_timeout)
    return None


#------------------------------------------------------------------------------


def connected(wait_timeout=5):
    ret = Deferred()

    if driver.is_enabled('service_proxy_transport'):
        p2p_connector_lookup = automat.find('p2p_connector')
        if p2p_connector_lookup:
            p2p_connector_machine = automat.by_index(p2p_connector_lookup[0])
            if p2p_connector_machine and p2p_connector_machine.state == 'CONNECTED':
                proxy_receiver_lookup = automat.find('proxy_receiver')
                if proxy_receiver_lookup:
                    proxy_receiver_machine = automat.by_index(proxy_receiver_lookup[0])
                    if proxy_receiver_machine and proxy_receiver_machine.state == 'LISTEN':
                        # service_proxy_transport() is enabled, proxy_receiver() is listening: all good
                        wait_timeout_defer = Deferred()
                        wait_timeout_defer.addBoth(
                            lambda _: ret.
                            callback({
                                'service_network': 'started',
                                'service_gateway': 'started',
                                'service_p2p_hookups': 'started',
                                'service_proxy_transport': 'started',
                                'proxy_receiver_state': proxy_receiver_machine.state,
                            }),
                        )
                        if not wait_timeout:
                            wait_timeout = 0.01
                        wait_timeout_defer.addTimeout(wait_timeout, clock=reactor)
                        return ret
                else:
                    d = driver.is_healthy('service_proxy_transport')
                    d.addCallback(on_service_proxy_transport_check_healthy, wait_timeout=wait_timeout)
                    d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='network_service.connected', ignore=True)
                    lg.warn('disconnected, reason is proxy_receiver() not started yet')
                    ret.callback(dict(
                        error='disconnected',
                        reason='proxy_receiver_not_started',
                    ))
                    return ret

    if not my_id.isLocalIdentityReady():
        lg.warn('local identity is not valid or not exist')
        ret.callback(dict(
            error='local identity is not valid or not exist',
            reason='identity_not_exist',
        ))
        return ret
    if not driver.is_enabled('service_network'):
        lg.warn('service_network() is disabled')
        ret.callback(dict(
            error='service_network() is disabled',
            reason='service_network_disabled',
        ))
        return ret
    if not driver.is_enabled('service_gateway'):
        lg.warn('service_gateway() is disabled')
        ret.callback(dict(
            error='service_gateway() is disabled',
            reason='service_gateway_disabled',
        ))
        return ret
    if not driver.is_enabled('service_p2p_hookups'):
        lg.warn('service_p2p_hookups() is disabled')
        ret.callback(dict(
            error='service_p2p_hookups() is disabled',
            reason='service_p2p_hookups_disabled',
        ))
        return ret

    do_service_test('service_network', ret, wait_timeout)
    return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
from .index import Index, IndexException, DocIdNotFound, ElemNotFound
import struct
import marshal
import os
import io
import shutil
from .storage import IU_Storage

from CodernityDB.env import cdb_environment
from CodernityDB.index import TryReindexException
import six

if cdb_environment.get('rlock_obj'):
    from CodernityDB import patch
    patch.patch_cache_rr(cdb_environment['rlock_obj'])

from CodernityDB.rr_cache import cache1lvl, cache2lvl

tree_buffer_size = io.DEFAULT_BUFFER_SIZE

cdb_environment['tree_buffer_size'] = tree_buffer_size


MODE_FIRST = 0
MODE_LAST = 1

MOVE_BUFFER_PREV = 0
MOVE_BUFFER_NEXT = 1


class NodeCapacityException(IndexException):
    pass


class IU_TreeBasedIndex(Index):

    custom_header = 'from CodernityDB.tree_index import TreeBasedIndex'

    def __init__(self, db_path, name, key_format='32s', pointer_format='I',
                 meta_format='32sIIc', node_capacity=10, storage_class=None):
        if node_capacity < 3:
            raise NodeCapacityException
        super(IU_TreeBasedIndex, self).__init__(db_path, name)
        self.data_start = self._start_ind + 1
        self.node_capacity = node_capacity
        self.flag_format = 'c'
        self.elements_counter_format = 'h'
        self.pointer_format = pointer_format
        self.key_format = key_format
        self.meta_format = meta_format
        self._count_props()
        if not storage_class:
            storage_class = IU_Storage
        if storage_class and not isinstance(storage_class, six.string_types):
            storage_class = storage_class.__name__
        self.storage_class = storage_class
        self.storage = None
        cache = cache1lvl(100)
        twolvl_cache = cache2lvl(150)
        self._find_key = cache(self._find_key)
        self._match_doc_id = cache(self._match_doc_id)
# self._read_single_leaf_record =
# twolvl_cache(self._read_single_leaf_record)
        self._find_key_in_leaf = twolvl_cache(self._find_key_in_leaf)
        self._read_single_node_key = twolvl_cache(self._read_single_node_key)
        self._find_first_key_occurence_in_node = twolvl_cache(
            self._find_first_key_occurence_in_node)
        self._find_last_key_occurence_in_node = twolvl_cache(
            self._find_last_key_occurence_in_node)
        self._read_leaf_nr_of_elements = cache(self._read_leaf_nr_of_elements)
        self._read_leaf_neighbours = cache(self._read_leaf_neighbours)
        self._read_leaf_nr_of_elements_and_neighbours = cache(
            self._read_leaf_nr_of_elements_and_neighbours)
        self._read_node_nr_of_elements_and_children_flag = cache(
            self._read_node_nr_of_elements_and_children_flag)

    def _count_props(self):
        """
        Counts dynamic properties for tree, such as all complex formats
        """
        self.single_leaf_record_format = self.key_format + self.meta_format
        self.single_node_record_format = self.pointer_format + \
            self.key_format + self.pointer_format
        self.node_format = self.elements_counter_format + self.flag_format\
            + self.pointer_format + (self.key_format +
                                     self.pointer_format) * self.node_capacity
        self.leaf_format = self.elements_counter_format + self.pointer_format * 2\
            + (self.single_leaf_record_format) * self.node_capacity
        self.leaf_heading_format = self.elements_counter_format + \
            self.pointer_format * 2
        self.node_heading_format = self.elements_counter_format + \
            self.flag_format
        self.key_size = struct.calcsize('<' + self.key_format)
        self.meta_size = struct.calcsize('<' + self.meta_format)
        self.single_leaf_record_size = struct.calcsize('<' + self.
                                                       single_leaf_record_format)
        self.single_node_record_size = struct.calcsize('<' + self.
                                                       single_node_record_format)
        self.node_size = struct.calcsize('<' + self.node_format)
        self.leaf_size = struct.calcsize('<' + self.leaf_format)
        self.flag_size = struct.calcsize('<' + self.flag_format)
        self.elements_counter_size = struct.calcsize('<' + self.
                                                     elements_counter_format)
        self.pointer_size = struct.calcsize('<' + self.pointer_format)
        self.leaf_heading_size = struct.calcsize(
            '<' + self.leaf_heading_format)
        self.node_heading_size = struct.calcsize(
            '<' + self.node_heading_format)

    def create_index(self):
        if os.path.isfile(os.path.join(self.db_path, self.name + '_buck')):
            raise IndexException('Already exists')
        with io.open(os.path.join(self.db_path, self.name + "_buck"), 'w+b') as f:
            props = dict(name=self.name,
                         flag_format=self.flag_format,
                         pointer_format=self.pointer_format,
                         elements_counter_format=self.elements_counter_format,
                         node_capacity=self.node_capacity,
                         key_format=self.key_format,
                         meta_format=self.meta_format,
                         version=self.__version__,
                         storage_class=self.storage_class)
            f.write(marshal.dumps(props))
        self.buckets = io.open(os.path.join(self.db_path, self.name +
                                            "_buck"), 'r+b', buffering=0)
        self._create_storage()
        self.buckets.seek(self._start_ind)
        self.buckets.write(struct.pack('<c', 'l'))
        self._insert_empty_root()
        self.root_flag = 'l'

    def destroy(self):
        super(IU_TreeBasedIndex, self).destroy()
        self._clear_cache()

    def open_index(self):
        if not os.path.isfile(os.path.join(self.db_path, self.name + '_buck')):
            raise IndexException("Doesn't exists")
        self.buckets = io.open(
            os.path.join(self.db_path, self.name + "_buck"), 'r+b', buffering=0)
        self.buckets.seek(self._start_ind)
        self.root_flag = struct.unpack('<c', self.buckets.read(1))[0]
        self._fix_params()
        self._open_storage()

    def _insert_empty_root(self):
        self.buckets.seek(self.data_start)
        root = struct.pack('<' + self.leaf_heading_format,
                           0,
                           0,
                           0)
        root += self.single_leaf_record_size * self.node_capacity * '\x00'
        self.buckets.write(root)
        self.flush()

    def insert(self, doc_id, key, start, size, status='o'):
        nodes_stack, indexes = self._find_leaf_to_insert(key)
        self._insert_new_record_into_leaf(nodes_stack.pop(),
                                          key,
                                          doc_id,
                                          start,
                                          size,
                                          status,
                                          nodes_stack,
                                          indexes)

        self._match_doc_id.delete(doc_id)

    def _read_leaf_nr_of_elements_and_neighbours(self, leaf_start):
        self.buckets.seek(leaf_start)
        data = self.buckets.read(
            self.elements_counter_size + 2 * self.pointer_size)
        nr_of_elements, prev_l, next_l = struct.unpack(
            '<' + self.elements_counter_format + 2 * self.pointer_format,
            data)
        return nr_of_elements, prev_l, next_l

    def _read_node_nr_of_elements_and_children_flag(self, start):
        self.buckets.seek(start)
        data = self.buckets.read(self.elements_counter_size + self.flag_size)
        nr_of_elements, children_flag = struct.unpack(
            '<' + self.elements_counter_format + self.flag_format,
            data)
        return nr_of_elements, children_flag

    def _read_leaf_nr_of_elements(self, start):
        self.buckets.seek(start)
        data = self.buckets.read(self.elements_counter_size)
        nr_of_elements = struct.unpack(
            '<' + self.elements_counter_format, data)
        return nr_of_elements[0]

    def _read_single_node_key(self, node_start, key_index):
        self.buckets.seek(self._calculate_key_position(
            node_start, key_index, 'n'))
        data = self.buckets.read(self.single_node_record_size)
        flag_left, key, pointer_right = struct.unpack(
            '<' + self.single_node_record_format, data)
        return flag_left, key, pointer_right

    def _read_single_leaf_record(self, leaf_start, key_index):
        self.buckets.seek(self._calculate_key_position(
            leaf_start, key_index, 'l'))
        data = self.buckets.read(self.single_leaf_record_size)
        key, doc_id, start, size, status = struct.unpack('<' + self.
                                                         single_leaf_record_format, data)
        return key, doc_id, start, size, status

    def _calculate_key_position(self, start, key_index, flag):
        """
        Calculates position of key in buckets file
        """
        if flag == 'l':
            return start + self.leaf_heading_size + key_index * self.single_leaf_record_size
        elif flag == 'n':
#            returns start position of flag before key[key_index]
            return start + self.node_heading_size + key_index * (self.pointer_size + self.key_size)

    def _match_doc_id(self, doc_id, key, element_index, leaf_start, nr_of_elements):
        curr_key_index = element_index + 1
        curr_leaf_start = leaf_start
        next_leaf = self._read_leaf_neighbours(leaf_start)[1]
        while True:
            if curr_key_index < nr_of_elements:
                curr_key, curr_doc_id, curr_start, curr_size,\
                    curr_status = self._read_single_leaf_record(
                        curr_leaf_start, curr_key_index)
                if key != curr_key:
#                    should't happen, crashes earlier on id index
                    raise DocIdNotFound
                elif doc_id == curr_doc_id and curr_status != 'd':
                    return curr_leaf_start, nr_of_elements, curr_key_index
                else:
                    curr_key_index = curr_key_index + 1
            else:  # there are no more elements in current leaf, must jump to next
                if not next_leaf:  # end of leaf linked list
#                    should't happen, crashes earlier on id index
                    raise DocIdNotFound
                else:
                    curr_leaf_start = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                    curr_key_index = 0

    def _find_existing(self, key, element_index, leaf_start, nr_of_elements):
        curr_key_index = element_index + 1
        curr_leaf_start = leaf_start
        next_leaf = self._read_leaf_neighbours(leaf_start)[1]
        while True:
            if curr_key_index < nr_of_elements:
                curr_key, curr_doc_id, curr_start, curr_size,\
                    curr_status = self._read_single_leaf_record(
                        curr_leaf_start, curr_key_index)
                if key != curr_key:
                    raise ElemNotFound
                elif curr_status != 'd':
                    return curr_leaf_start, nr_of_elements, curr_key_index
                else:
                    curr_key_index = curr_key_index + 1
            else:  # there are no more elements in current leaf, must jump to next
                if not next_leaf:  # end of leaf linked list
                    raise ElemNotFound
                else:
                    curr_leaf_start = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                    curr_key_index = 0

    def _update_element(self, leaf_start, key_index, new_data):
        self.buckets.seek(self._calculate_key_position(leaf_start, key_index, 'l')
                          + self.key_size)
        self.buckets.write(struct.pack('<' + self.meta_format,
                                       *new_data))

#        self._read_single_leaf_record.delete(leaf_start_position, key_index)

    def _delete_element(self, leaf_start, key_index):
        self.buckets.seek(self._calculate_key_position(leaf_start, key_index, 'l')
                          + self.single_leaf_record_size - 1)
        self.buckets.write(struct.pack('<c', 'd'))

#        self._read_single_leaf_record.delete(leaf_start_position, key_index)

    def _leaf_linear_key_search(self, key, start, start_index, end_index):
        self.buckets.seek(start)
        data = self.buckets.read(
            (end_index - start_index + 1) * self.single_leaf_record_size)
        curr_key = struct.unpack(
            '<' + self.key_format, data[:self.key_size])[0]
        data = data[self.single_leaf_record_size:]
        curr_index = 0
        while curr_key != key:
            curr_index += 1
            curr_key = struct.unpack(
                '<' + self.key_format, data[:self.key_size])[0]
            data = data[self.single_leaf_record_size:]
        return start_index + curr_index

    def _node_linear_key_search(self, key, start, start_index, end_index):
        self.buckets.seek(start + self.pointer_size)
        data = self.buckets.read((end_index - start_index + 1) * (
            self.key_size + self.pointer_size))
        curr_key = struct.unpack(
            '<' + self.key_format, data[:self.key_size])[0]
        data = data[self.key_size + self.pointer_size:]
        curr_index = 0
        while curr_key != key:
            curr_index += 1
            curr_key = struct.unpack(
                '<' + self.key_format, data[:self.key_size])[0]
            data = data[self.key_size + self.pointer_size:]
        return start_index + curr_index

    def _next_buffer(self, buffer_start, buffer_end):
            return buffer_end, buffer_end + tree_buffer_size

    def _prev_buffer(self, buffer_start, buffer_end):
            return buffer_start - tree_buffer_size, buffer_start

    def _choose_next_candidate_index_in_leaf(self, leaf_start, candidate_start, buffer_start, buffer_end, imin, imax):
        if buffer_start > candidate_start:
            move_buffer = MOVE_BUFFER_PREV
        elif buffer_end < candidate_start + self.single_leaf_record_size:
            move_buffer = MOVE_BUFFER_NEXT
        else:
            move_buffer = None
        return self._calculate_key_position(leaf_start, (imin + imax) / 2, 'l'), (imin + imax) / 2, move_buffer

    def _choose_next_candidate_index_in_node(self, node_start, candidate_start, buffer_start, buffer_end, imin, imax):
        if buffer_start > candidate_start:
            move_buffer = MOVE_BUFFER_PREV
        elif buffer_end < candidate_start + self.single_node_record_size:
            (self.pointer_size + self.key_size) - 1
            move_buffer = MOVE_BUFFER_NEXT
        else:
            move_buffer = None
        return self._calculate_key_position(node_start, (imin + imax) / 2, 'n'), (imin + imax) / 2, move_buffer

    def _find_key_in_leaf(self, leaf_start, key, nr_of_elements):
        if nr_of_elements == 1:
            return self._find_key_in_leaf_with_one_element(key, leaf_start)[-5:]
        else:
            return self._find_key_in_leaf_using_binary_search(key, leaf_start, nr_of_elements)[-5:]

    def _find_key_in_leaf_for_update(self, key, doc_id, leaf_start, nr_of_elements):
        if nr_of_elements == 1:
            return self._find_key_in_leaf_with_one_element(key, leaf_start, doc_id=doc_id)
        else:
            return self._find_key_in_leaf_using_binary_search(key, leaf_start, nr_of_elements, mode=MODE_FIRST, doc_id=doc_id)

    def _find_index_of_first_key_equal_or_smaller_key(self, key, leaf_start, nr_of_elements):
        if nr_of_elements == 1:
            return self._find_key_in_leaf_with_one_element(key, leaf_start, mode=MODE_FIRST, return_closest=True)[:2]
        else:
            return self._find_key_in_leaf_using_binary_search(key, leaf_start, nr_of_elements, mode=MODE_FIRST, return_closest=True)[:2]

    def _find_index_of_last_key_equal_or_smaller_key(self, key, leaf_start, nr_of_elements):
        if nr_of_elements == 1:
            return self._find_key_in_leaf_with_one_element(key, leaf_start, mode=MODE_LAST, return_closest=True)[:2]
        else:
            return self._find_key_in_leaf_using_binary_search(key, leaf_start, nr_of_elements, mode=MODE_LAST, return_closest=True)[:2]

    def _find_index_of_first_key_equal(self, key, leaf_start, nr_of_elements):
        if nr_of_elements == 1:
            return self._find_key_in_leaf_with_one_element(key, leaf_start, mode=MODE_FIRST)[:2]
        else:
            return self._find_key_in_leaf_using_binary_search(key, leaf_start, nr_of_elements, mode=MODE_FIRST)[:2]

    def _find_key_in_leaf_with_one_element(self, key, leaf_start, doc_id=None, mode=None, return_closest=False):
        curr_key, curr_doc_id, curr_start, curr_size,\
            curr_status = self._read_single_leaf_record(leaf_start, 0)
        if key != curr_key:
            if return_closest and curr_status != 'd':
                return leaf_start, 0
            else:
                raise ElemNotFound
        else:
            if curr_status == 'd':
                raise ElemNotFound
            elif doc_id is not None and doc_id != curr_doc_id:
#                    should't happen, crashes earlier on id index
                raise DocIdNotFound
            else:
                return leaf_start, 0, curr_doc_id, curr_key, curr_start, curr_size, curr_status

    def _find_key_in_leaf_using_binary_search(self, key, leaf_start, nr_of_elements, doc_id=None, mode=None, return_closest=False):
        """
        Binary search implementation used in all get functions
        """
        imin, imax = 0, nr_of_elements - 1
        buffer_start, buffer_end = self._set_buffer_limits()
        candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_leaf(leaf_start,
                                                                                                  self._calculate_key_position(leaf_start,
                                                                                                                               (imin + imax) / 2,
                                                                                                                               'l'),
                                                                                                  buffer_start,
                                                                                                  buffer_end,
                                                                                                  imin, imax)
        while imax != imin and imax > imin:
            curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                      candidate_index)
            candidate_start = self._calculate_key_position(
                leaf_start, candidate_index, 'l')
            if key < curr_key:
                if move_buffer == MOVE_BUFFER_PREV:
                    buffer_start, buffer_end = self._prev_buffer(
                        buffer_start, buffer_end)
                else:  # if next chosen element is in current buffer, abort moving to other
                    move_buffer is None
                imax = candidate_index - 1
                candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_leaf(leaf_start,
                                                                                                          candidate_start,
                                                                                                          buffer_start,
                                                                                                          buffer_end,
                                                                                                          imin, imax)
            elif key == curr_key:
                if mode == MODE_LAST:
                    if move_buffer == MOVE_BUFFER_NEXT:
                        buffer_start, buffer_end = self._next_buffer(
                            buffer_start, buffer_end)
                    else:
                        move_buffer is None
                    imin = candidate_index + 1
                    candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_leaf(leaf_start,
                                                                                                              candidate_start,
                                                                                                              buffer_start,
                                                                                                              buffer_end,
                                                                                                              imin, imax)
                else:
                    if curr_status == 'o':
                        break
                    else:
                        if move_buffer == MOVE_BUFFER_PREV:
                            buffer_start, buffer_end = self._prev_buffer(
                                buffer_start, buffer_end)
                        else:
                            move_buffer is None
                        imax = candidate_index
                        candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_leaf(leaf_start,
                                                                                                                  candidate_start,
                                                                                                                  buffer_start,
                                                                                                                  buffer_end,
                                                                                                                  imin, imax)
            else:
                if move_buffer == MOVE_BUFFER_NEXT:
                    buffer_start, buffer_end = self._next_buffer(
                        buffer_start, buffer_end)
                else:
                    move_buffer is None
                imin = candidate_index + 1
                candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_leaf(leaf_start,
                                                                                                          candidate_start,
                                                                                                          buffer_start,
                                                                                                          buffer_end,
                                                                                                          imin, imax)

        if imax > imin:
            chosen_key_position = candidate_index
        else:
            chosen_key_position = imax
        curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                  chosen_key_position)
        if key != curr_key:
            if return_closest:  # useful for find all bigger/smaller methods
                return leaf_start, chosen_key_position
            else:
                raise ElemNotFound
        if doc_id and doc_id == curr_doc_id and curr_status == 'o':
            return leaf_start, chosen_key_position, curr_doc_id, curr_key, curr_start, curr_size, curr_status
        else:
            if mode == MODE_FIRST and imin < chosen_key_position:  # check if there isn't any element with equal key before chosen one
                matching_record_index = self._leaf_linear_key_search(key,
                                                                     self._calculate_key_position(leaf_start,
                                                                                                  imin,
                                                                                                  'l'),
                                                                     imin,
                                                                     chosen_key_position)
            else:
                matching_record_index = chosen_key_position
            curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                      matching_record_index)
            if curr_status == 'd' and not return_closest:
                leaf_start, nr_of_elements, matching_record_index = self._find_existing(key,
                                                                                        matching_record_index,
                                                                                        leaf_start,
                                                                                        nr_of_elements)
                curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                          matching_record_index)
            if doc_id is not None and doc_id != curr_doc_id:
                leaf_start, nr_of_elements, matching_record_index = self._match_doc_id(doc_id,
                                                                                       key,
                                                                                       matching_record_index,
                                                                                       leaf_start,
                                                                                       nr_of_elements)
                curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                          matching_record_index)
            return leaf_start, matching_record_index, curr_doc_id, curr_key, curr_start, curr_size, curr_status

    def _find_place_in_leaf(self, key, leaf_start, nr_of_elements):
        if nr_of_elements == 1:
            return self._find_place_in_leaf_with_one_element(key, leaf_start)
        else:
            return self._find_place_in_leaf_using_binary_search(key, leaf_start, nr_of_elements)

    def _find_place_in_leaf_with_one_element(self, key, leaf_start):
        curr_key, curr_doc_id, curr_start, curr_size,\
            curr_status = self._read_single_leaf_record(leaf_start, 0)
        if curr_status == 'd':
                return leaf_start, 0, 0, False, True  # leaf start, index of new key position, nr of rec to rewrite, full_leaf flag, on_deleted flag
        else:
            if key < curr_key:
                return leaf_start, 0, 1, False, False
            else:
                return leaf_start, 1, 0, False, False

    def _find_place_in_leaf_using_binary_search(self, key, leaf_start, nr_of_elements):
        """
        Binary search implementation used in insert function
        """
        imin, imax = 0, nr_of_elements - 1
        buffer_start, buffer_end = self._set_buffer_limits()
        candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_leaf(leaf_start,
                                                                                                  self._calculate_key_position(leaf_start,
                                                                                                                               (imin + imax) / 2,
                                                                                                                               'l'),
                                                                                                  buffer_start,
                                                                                                  buffer_end,
                                                                                                  imin, imax)
        while imax != imin and imax > imin:
            curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                      candidate_index)
            candidate_start = self._calculate_key_position(
                leaf_start, candidate_index, 'l')
            if key < curr_key:
                if move_buffer == MOVE_BUFFER_PREV:
                    buffer_start, buffer_end = self._prev_buffer(
                        buffer_start, buffer_end)
                else:  # if next chosen element is in current buffer, abort moving to other
                    move_buffer is None
                imax = candidate_index - 1
                candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_leaf(leaf_start,
                                                                                                          candidate_start,
                                                                                                          buffer_start,
                                                                                                          buffer_end,
                                                                                                          imin, imax)
            else:
                if move_buffer == MOVE_BUFFER_NEXT:
                    buffer_start, buffer_end = self._next_buffer(
                        buffer_start, buffer_end)
                else:
                    move_buffer is None
                imin = candidate_index + 1
                candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_leaf(leaf_start,
                                                                                                          candidate_start,
                                                                                                          buffer_start,
                                                                                                          buffer_end,
                                                                                                          imin, imax)
        if imax < imin and imin < nr_of_elements:
            chosen_key_position = imin
        else:
            chosen_key_position = imax
        curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                  chosen_key_position)
        if curr_status == 'd':
            return leaf_start, chosen_key_position, 0, False, True
        elif key < curr_key:
            if chosen_key_position > 0:
                curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                          chosen_key_position - 1)
                if curr_start == 'd':
                    return leaf_start, chosen_key_position - 1, 0, False, True
                else:
                    return leaf_start, chosen_key_position, nr_of_elements - chosen_key_position, (nr_of_elements == self.node_capacity), False
            else:
                return leaf_start, chosen_key_position, nr_of_elements - chosen_key_position, (nr_of_elements == self.node_capacity), False
        else:
            if chosen_key_position < nr_of_elements - 1:
                curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_start,
                                                                                                          chosen_key_position + 1)
                if curr_start == 'd':
                    return leaf_start, chosen_key_position + 1, 0, False, True
                else:
                    return leaf_start, chosen_key_position + 1, nr_of_elements - chosen_key_position - 1, (nr_of_elements == self.node_capacity), False
            else:
                return leaf_start, chosen_key_position + 1, nr_of_elements - chosen_key_position - 1, (nr_of_elements == self.node_capacity), False

    def _set_buffer_limits(self):
        pos = self.buckets.tell()
        buffer_start = pos - (pos % tree_buffer_size)
        return buffer_start, (buffer_start + tree_buffer_size)

    def _find_first_key_occurence_in_node(self, node_start, key, nr_of_elements):
        if nr_of_elements == 1:
            return self._find_key_in_node_with_one_element(key, node_start, mode=MODE_FIRST)
        else:
            return self._find_key_in_node_using_binary_search(key, node_start, nr_of_elements, mode=MODE_FIRST)

    def _find_last_key_occurence_in_node(self, node_start, key, nr_of_elements):
        if nr_of_elements == 1:
            return self._find_key_in_node_with_one_element(key, node_start, mode=MODE_LAST)
        else:
            return self._find_key_in_node_using_binary_search(key, node_start, nr_of_elements, mode=MODE_LAST)

    def _find_key_in_node_with_one_element(self, key, node_start, mode=None):
        l_pointer, curr_key, r_pointer = self._read_single_node_key(
            node_start, 0)
        if key < curr_key:
            return 0, l_pointer
        elif key > curr_key:
            return 0, r_pointer
        else:
            if mode == MODE_FIRST:
                return 0, l_pointer
            elif mode == MODE_LAST:
                return 0, r_pointer
            else:
                raise Exception('Invalid mode declared: set first/last')

    def _find_key_in_node_using_binary_search(self, key, node_start, nr_of_elements, mode=None):
        imin, imax = 0, nr_of_elements - 1
        buffer_start, buffer_end = self._set_buffer_limits()
        candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_node(node_start,
                                                                                                  self._calculate_key_position(node_start,
                                                                                                                               (imin + imax) / 2,
                                                                                                                               'n'),
                                                                                                  buffer_start,
                                                                                                  buffer_end,
                                                                                                  imin, imax)
        while imax != imin and imax > imin:
            l_pointer, curr_key, r_pointer = self._read_single_node_key(
                node_start, candidate_index)
            candidate_start = self._calculate_key_position(
                node_start, candidate_index, 'n')
            if key < curr_key:
                if move_buffer == MOVE_BUFFER_PREV:
                    buffer_start, buffer_end = self._prev_buffer(
                        buffer_start, buffer_end)
                else:  # if next chosen element is in current buffer, abort moving to other
                    move_buffer is None
                imax = candidate_index - 1
                candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_node(node_start,
                                                                                                          candidate_start,
                                                                                                          buffer_start,
                                                                                                          buffer_end,
                                                                                                          imin, imax)
            elif key == curr_key:
                if mode == MODE_LAST:
                    if move_buffer == MOVE_BUFFER_NEXT:
                        buffer_start, buffer_end = self._next_buffer(
                            buffer_start, buffer_end)
                    else:
                        move_buffer is None
                    imin = candidate_index + 1
                    candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_node(node_start,
                                                                                                              candidate_start,
                                                                                                              buffer_start,
                                                                                                              buffer_end,
                                                                                                              imin, imax)
                else:
                    break
            else:
                if move_buffer == MOVE_BUFFER_NEXT:
                        buffer_start, buffer_end = self._next_buffer(
                            buffer_start, buffer_end)
                else:
                    move_buffer is None
                imin = candidate_index + 1
                candidate_start, candidate_index, move_buffer = self._choose_next_candidate_index_in_node(node_start,
                                                                                                          candidate_start,
                                                                                                          buffer_start,
                                                                                                          buffer_end,
                                                                                                          imin, imax)

        if imax > imin:
            chosen_key_position = candidate_index
        elif imax < imin and imin < nr_of_elements:
            chosen_key_position = imin
        else:
            chosen_key_position = imax
        l_pointer, curr_key, r_pointer = self._read_single_node_key(
            node_start, chosen_key_position)
        if mode == MODE_FIRST and imin < chosen_key_position:  # check if there is no elements with equal key before chosen one
            matching_record_index = self._node_linear_key_search(key,
                                                                 self._calculate_key_position(node_start,
                                                                                              imin,
                                                                                              'n'),
                                                                 imin,
                                                                 chosen_key_position)
        else:
            matching_record_index = chosen_key_position
        l_pointer, curr_key, r_pointer = self._read_single_node_key(
            node_start, matching_record_index)
        if key < curr_key:
            return matching_record_index, l_pointer
        elif key > curr_key:
            return matching_record_index, r_pointer
        else:
            if mode == MODE_FIRST:
                return matching_record_index, l_pointer
            elif mode == MODE_LAST:
                return matching_record_index, r_pointer
            else:
                raise Exception('Invalid mode declared: first/last')

    def _update_leaf_ready_data(self, leaf_start, start_index, new_nr_of_elements, records_to_rewrite):
        self.buckets.seek(leaf_start)
        self.buckets.write(struct.pack('<h', new_nr_of_elements))
        start_position = self._calculate_key_position(
            leaf_start, start_index, 'l')
        self.buckets.seek(start_position)
        self.buckets.write(
            struct.pack(
                '<' + (new_nr_of_elements - start_index) *
                self.single_leaf_record_format,
                *records_to_rewrite))

#        self._read_single_leaf_record.delete(leaf_start)
        self._read_leaf_nr_of_elements.delete(leaf_start)
        self._read_leaf_nr_of_elements_and_neighbours.delete(leaf_start)

    def _update_leaf(self, leaf_start, new_record_position, nr_of_elements,
                     nr_of_records_to_rewrite, on_deleted, new_key,
                     new_doc_id, new_start, new_size, new_status):
        if nr_of_records_to_rewrite == 0:  # just write at set position
            self.buckets.seek(self._calculate_key_position(
                leaf_start, new_record_position, 'l'))
            self.buckets.write(
                struct.pack('<' + self.single_leaf_record_format,
                            new_key,
                            new_doc_id,
                            new_start,
                            new_size,
                            new_status))
            self.flush()
        else:  # must read all elems after new one, and rewrite them after new
            start = self._calculate_key_position(
                leaf_start, new_record_position, 'l')
            self.buckets.seek(start)
            data = self.buckets.read(nr_of_records_to_rewrite *
                                     self.single_leaf_record_size)
            records_to_rewrite = struct.unpack('<' + nr_of_records_to_rewrite *
                                               self.single_leaf_record_format, data)
            curr_index = 0
            records_to_rewrite = list(records_to_rewrite)
            for status in records_to_rewrite[4::5]:  # don't write back deleted records, deleting them from list
                if status != 'o':
                    del records_to_rewrite[curr_index * 5:curr_index * 5 + 5]
                    nr_of_records_to_rewrite -= 1
                    nr_of_elements -= 1
                else:
                    curr_index += 1

            self.buckets.seek(start)
            self.buckets.write(
                struct.pack(
                    '<' + (nr_of_records_to_rewrite +
                           1) * self.single_leaf_record_format,
                    new_key,
                    new_doc_id,
                    new_start,
                    new_size,
                    new_status,
                    *tuple(records_to_rewrite)))
            self.flush()
        self.buckets.seek(leaf_start)
        if not on_deleted:  # when new record replaced deleted one, nr of leaf elements stays the same
            self.buckets.write(struct.pack('<h', nr_of_elements + 1))

        self._read_leaf_nr_of_elements.delete(leaf_start)
        self._read_leaf_nr_of_elements_and_neighbours.delete(leaf_start)
        self._find_key_in_leaf.delete(leaf_start)
#        self._read_single_leaf_record.delete(leaf_start)

    def _read_leaf_neighbours(self, leaf_start):
        self.buckets.seek(leaf_start + self.elements_counter_size)
        neihbours_data = self.buckets.read(2 * self.pointer_size)
        prev_l, next_l = struct.unpack(
            '<' + 2 * self.pointer_format, neihbours_data)
        return prev_l, next_l

    def _update_leaf_size_and_pointers(self, leaf_start, new_size, new_prev, new_next):
        self.buckets.seek(leaf_start)
        self.buckets.write(
            struct.pack(
                '<' + self.elements_counter_format + 2 * self.pointer_format,
                new_size,
                new_prev,
                new_next))

        self._read_leaf_nr_of_elements.delete(leaf_start)
        self._read_leaf_neighbours.delete(leaf_start)
        self._read_leaf_nr_of_elements_and_neighbours.delete(leaf_start)

    def _update_leaf_prev_pointer(self, leaf_start, pointer):
        self.buckets.seek(leaf_start + self.elements_counter_size)
        self.buckets.write(struct.pack('<' + self.pointer_format,
                                       pointer))

        self._read_leaf_neighbours.delete(leaf_start)
        self._read_leaf_nr_of_elements_and_neighbours.delete(leaf_start)

    def _update_size(self, start, new_size):
        self.buckets.seek(start)
        self.buckets.write(struct.pack('<' + self.elements_counter_format,
                                       new_size))

        self._read_leaf_nr_of_elements.delete(start)
        self._read_leaf_nr_of_elements_and_neighbours.delete(start)

    def _create_new_root_from_leaf(self, leaf_start, nr_of_records_to_rewrite, new_leaf_size, old_leaf_size, half_size, new_data):
        blanks = (self.node_capacity - new_leaf_size) * \
            self.single_leaf_record_size * '\x00'
        left_leaf_start_position = self.data_start + self.node_size
        right_leaf_start_position = self.data_start + \
            self.node_size + self.leaf_size
        self.buckets.seek(self.data_start + self.leaf_heading_size)
        # read old root
        data = self.buckets.read(
            self.single_leaf_record_size * self.node_capacity)
        leaf_data = struct.unpack('<' + self.
                                  single_leaf_record_format * self.node_capacity, data)
        # remove deleted records, if succeded abort spliting
        if self._update_if_has_deleted(self.data_start, leaf_data, 0, new_data):
            return None
        # find out key which goes to parent node
        if nr_of_records_to_rewrite > new_leaf_size - 1:
            key_moved_to_parent_node = leaf_data[(old_leaf_size - 1) * 5]
        elif nr_of_records_to_rewrite == new_leaf_size - 1:
            key_moved_to_parent_node = new_data[0]
        else:
            key_moved_to_parent_node = leaf_data[old_leaf_size * 5]
        data_to_write = self._prepare_new_root_data(key_moved_to_parent_node,
                                                    left_leaf_start_position,
                                                    right_leaf_start_position,
                                                    'l')
        if nr_of_records_to_rewrite > half_size:
                # key goes to first half
                # prepare left leaf data
            left_leaf_data = struct.pack('<' + self.leaf_heading_format + self.single_leaf_record_format
                                         * (self.node_capacity - nr_of_records_to_rewrite),
                                         old_leaf_size,
                                         0,
                                         right_leaf_start_position,
                                         *leaf_data[:-nr_of_records_to_rewrite * 5])
            left_leaf_data += struct.pack(
                '<' + self.single_leaf_record_format * (
                    nr_of_records_to_rewrite - new_leaf_size + 1),
                new_data[0],
                new_data[1],
                new_data[2],
                new_data[3],
                new_data[4],
                *leaf_data[-nr_of_records_to_rewrite * 5:(old_leaf_size - 1) * 5])
                # prepare right leaf_data
            right_leaf_data = struct.pack('<' + self.elements_counter_format + 2 * self.pointer_format +
                                          self.single_leaf_record_format *
                                          new_leaf_size,
                                          new_leaf_size,
                                          left_leaf_start_position,
                                          0,
                                          *leaf_data[-new_leaf_size * 5:])
        else:
                # key goes to second half
            if nr_of_records_to_rewrite:
                records_before = leaf_data[old_leaf_size *
                                           5:-nr_of_records_to_rewrite * 5]
                records_after = leaf_data[-nr_of_records_to_rewrite * 5:]
            else:
                records_before = leaf_data[old_leaf_size * 5:]
                records_after = []

            left_leaf_data = struct.pack(
                '<' + self.leaf_heading_format +
                self.single_leaf_record_format * old_leaf_size,
                old_leaf_size,
                0,
                right_leaf_start_position,
                *leaf_data[:old_leaf_size * 5])
                # prepare right leaf_data
            right_leaf_data = struct.pack('<' + self.elements_counter_format + 2 * self.pointer_format +
                                          self.single_leaf_record_format * (new_leaf_size -
                                                                            nr_of_records_to_rewrite - 1),
                                          new_leaf_size,
                                          left_leaf_start_position,
                                          0,
                                          *records_before)
            right_leaf_data += struct.pack(
                '<' + self.single_leaf_record_format * (
                    nr_of_records_to_rewrite + 1),
                new_data[0],
                new_data[1],
                new_data[2],
                new_data[3],
                new_data[4],
                *records_after)
        left_leaf_data += (self.node_capacity -
                           old_leaf_size) * self.single_leaf_record_size * '\x00'
        right_leaf_data += blanks
        data_to_write += left_leaf_data
        data_to_write += right_leaf_data
        self.buckets.seek(self._start_ind)
        self.buckets.write(struct.pack('<c', 'n') + data_to_write)
        self.root_flag = 'n'

#            self._read_single_leaf_record.delete(leaf_start)
        self._find_key_in_leaf.delete(leaf_start)
        self._read_leaf_nr_of_elements.delete(leaf_start)
        self._read_leaf_nr_of_elements_and_neighbours.delete(leaf_start)
        self._read_leaf_neighbours.delete(leaf_start)
        return None

    def _split_leaf(
        self, leaf_start, nr_of_records_to_rewrite, new_key, new_doc_id, new_start, new_size, new_status,
            create_new_root=False):
        """
        Splits full leaf in two separate ones, first half of records stays on old position,
        second half is written as new leaf at the end of file.
        """
        half_size = self.node_capacity / 2
        if self.node_capacity % 2 == 0:
            old_leaf_size = half_size
            new_leaf_size = half_size + 1
        else:
            old_leaf_size = new_leaf_size = half_size + 1
        if create_new_root:  # leaf is a root
            new_data = [new_key, new_doc_id, new_start, new_size, new_status]
            self._create_new_root_from_leaf(leaf_start, nr_of_records_to_rewrite, new_leaf_size, old_leaf_size, half_size, new_data)
        else:
            blanks = (self.node_capacity - new_leaf_size) * \
                self.single_leaf_record_size * '\x00'
            prev_l, next_l = self._read_leaf_neighbours(leaf_start)
            if nr_of_records_to_rewrite > half_size:  # insert key into first half of leaf
                self.buckets.seek(self._calculate_key_position(leaf_start,
                                                               self.node_capacity - nr_of_records_to_rewrite,
                                                               'l'))
                # read all records with key>new_key
                data = self.buckets.read(
                    nr_of_records_to_rewrite * self.single_leaf_record_size)
                records_to_rewrite = struct.unpack(
                    '<' + nr_of_records_to_rewrite * self.single_leaf_record_format, data)
                # remove deleted records, if succeded abort spliting
                if self._update_if_has_deleted(leaf_start,
                                               records_to_rewrite,
                                               self.node_capacity -
                                               nr_of_records_to_rewrite,
                                               [new_key, new_doc_id, new_start, new_size, new_status]):
                    return None
                key_moved_to_parent_node = records_to_rewrite[
                    -new_leaf_size * 5]
                # write new leaf at end of file
                self.buckets.seek(0, 2)  # end of file
                new_leaf_start = self.buckets.tell()
                # prepare new leaf_data
                new_leaf = struct.pack('<' + self.elements_counter_format + 2 * self.pointer_format +
                                       self.single_leaf_record_format *
                                       new_leaf_size,
                                       new_leaf_size,
                                       leaf_start,
                                       next_l,
                                       *records_to_rewrite[-new_leaf_size * 5:])
                new_leaf += blanks
                # write new leaf
                self.buckets.write(new_leaf)
                # update old leaf heading
                self._update_leaf_size_and_pointers(leaf_start,
                                                    old_leaf_size,
                                                    prev_l,
                                                    new_leaf_start)
                # seek position of new key in first half
                self.buckets.seek(self._calculate_key_position(leaf_start,
                                                               self.node_capacity - nr_of_records_to_rewrite,
                                                               'l'))
                # write new key and keys after
                self.buckets.write(
                    struct.pack(
                        '<' + self.single_leaf_record_format *
                        (nr_of_records_to_rewrite - new_leaf_size + 1),
                        new_key,
                        new_doc_id,
                        new_start,
                        new_size,
                        'o',
                        *records_to_rewrite[:-new_leaf_size * 5]))

                if next_l:  # when next_l is 0 there is no next leaf to update, avoids writing data at 0 position of file
                    self._update_leaf_prev_pointer(
                        next_l, new_leaf_start)

#                self._read_single_leaf_record.delete(leaf_start)
                self._find_key_in_leaf.delete(leaf_start)

                return new_leaf_start, key_moved_to_parent_node
            else:  # key goes into second half of leaf     '
                # seek half of the leaf
                self.buckets.seek(self._calculate_key_position(
                    leaf_start, old_leaf_size, 'l'))
                data = self.buckets.read(
                    self.single_leaf_record_size * (new_leaf_size - 1))
                records_to_rewrite = struct.unpack('<' + (new_leaf_size - 1) *
                                                   self.single_leaf_record_format, data)
                # remove deleted records, if succeded abort spliting
                if self._update_if_has_deleted(leaf_start,
                                               records_to_rewrite,
                                               old_leaf_size,
                                               [new_key, new_doc_id, new_start, new_size, new_status]):
                    return None
                key_moved_to_parent_node = records_to_rewrite[
                    -(new_leaf_size - 1) * 5]
                if key_moved_to_parent_node > new_key:
                    key_moved_to_parent_node = new_key
                self.buckets.seek(0, 2)  # end of file
                new_leaf_start = self.buckets.tell()
                # prepare new leaf data
                index_of_records_split = nr_of_records_to_rewrite * 5
                if index_of_records_split:
                    records_before = records_to_rewrite[
                        :-index_of_records_split]
                    records_after = records_to_rewrite[
                        -index_of_records_split:]
                else:
                    records_before = records_to_rewrite
                    records_after = []
                new_leaf = struct.pack('<' + self.elements_counter_format + 2 * self.pointer_format
                                       + self.single_leaf_record_format * (new_leaf_size -
                                                                           nr_of_records_to_rewrite - 1),
                                       new_leaf_size,
                                       leaf_start,
                                       next_l,
                                       *records_before)
                new_leaf += struct.pack(
                    '<' + self.single_leaf_record_format *
                    (nr_of_records_to_rewrite + 1),
                    new_key,
                    new_doc_id,
                    new_start,
                    new_size,
                    'o',
                    *records_after)
                new_leaf += blanks
                self.buckets.write(new_leaf)
                self._update_leaf_size_and_pointers(leaf_start,
                                                    old_leaf_size,
                                                    prev_l,
                                                    new_leaf_start)
                if next_l:  # pren next_l is 0 there is no next leaf to update, avoids writing data at 0 position of file
                    self._update_leaf_prev_pointer(
                        next_l, new_leaf_start)

#                self._read_single_leaf_record.delete(leaf_start)
                self._find_key_in_leaf.delete(leaf_start)

                return new_leaf_start, key_moved_to_parent_node

    def _update_if_has_deleted(self, leaf_start, records_to_rewrite, start_position, new_record_data):
        """
        Checks if there are any deleted elements in data to rewrite and prevent from writing then back.
        """
        curr_index = 0
        nr_of_elements = self.node_capacity
        records_to_rewrite = list(records_to_rewrite)
        for status in records_to_rewrite[4::5]:  # remove deleted from list
            if status != 'o':
                del records_to_rewrite[curr_index * 5:curr_index * 5 + 5]
                nr_of_elements -= 1
            else:
                curr_index += 1
        # if were deleted dont have to split, just update leaf
        if nr_of_elements < self.node_capacity:
            data_split_index = 0
            for key in records_to_rewrite[0::5]:
                if key > new_record_data[0]:
                    break
                else:
                    data_split_index += 1
            records_to_rewrite = records_to_rewrite[:data_split_index * 5]\
                + new_record_data\
                + records_to_rewrite[data_split_index * 5:]
            self._update_leaf_ready_data(leaf_start,
                                         start_position,
                                         nr_of_elements + 1,
                                         records_to_rewrite),
            return True
        else:  # did not found any deleted records in leaf
            return False

    def _prepare_new_root_data(self, root_key, left_pointer, right_pointer, children_flag='n'):
        new_root = struct.pack(
            '<' + self.node_heading_format + self.single_node_record_format,
            1,
            children_flag,
            left_pointer,
            root_key,
            right_pointer)
        new_root += (self.key_size + self.pointer_size) * (self.
                                                           node_capacity - 1) * '\x00'
        return new_root

    def _create_new_root_from_node(self, node_start, children_flag, nr_of_keys_to_rewrite, new_node_size, old_node_size, new_key, new_pointer):
            # reading second half of node
            self.buckets.seek(self.data_start + self.node_heading_size)
            # read all keys with key>new_key
            data = self.buckets.read(self.pointer_size + self.
                                     node_capacity * (self.key_size + self.pointer_size))
            old_node_data = struct.unpack('<' + self.pointer_format + self.node_capacity *
                                          (self.key_format + self.pointer_format), data)
            self.buckets.seek(0, 2)  # end of file
            new_node_start = self.buckets.tell()
            if nr_of_keys_to_rewrite == new_node_size:
                key_moved_to_root = new_key
                # prepare new nodes data
                left_node = struct.pack('<' + self.node_heading_format + self.pointer_format +
                                        old_node_size * (self.
                                                         key_format + self.pointer_format),
                                        old_node_size,
                                        children_flag,
                                        *old_node_data[:old_node_size * 2 + 1])

                right_node = struct.pack('<' + self.node_heading_format + self.pointer_format +
                                         new_node_size * (self.
                                                          key_format + self.pointer_format),
                                         new_node_size,
                                         children_flag,
                                         new_pointer,
                                         *old_node_data[old_node_size * 2 + 1:])
            elif nr_of_keys_to_rewrite > new_node_size:
                key_moved_to_root = old_node_data[old_node_size * 2 - 1]
                # prepare new nodes data
                if nr_of_keys_to_rewrite == self.node_capacity:
                    keys_before = old_node_data[:1]
                    keys_after = old_node_data[1:old_node_size * 2 - 1]
                else:
                    keys_before = old_node_data[:-nr_of_keys_to_rewrite * 2]
                    keys_after = old_node_data[-(
                        nr_of_keys_to_rewrite) * 2:old_node_size * 2 - 1]
                left_node = struct.pack('<' + self.node_heading_format + self.pointer_format +
                                        (self.node_capacity - nr_of_keys_to_rewrite) * (self.
                                                                                        key_format + self.pointer_format),
                                        old_node_size,
                                        children_flag,
                                        *keys_before)
                left_node += struct.pack(
                    '<' + (self.key_format + self.pointer_format) *
                    (nr_of_keys_to_rewrite - new_node_size),
                    new_key,
                    new_pointer,
                    *keys_after)

                right_node = struct.pack('<' + self.node_heading_format + self.pointer_format +
                                         new_node_size * (self.
                                                          key_format + self.pointer_format),
                                         new_node_size,
                                         children_flag,
                                         *old_node_data[old_node_size * 2:])
            else:
#               'inserting key into second half of node and creating new root'
                key_moved_to_root = old_node_data[old_node_size * 2 + 1]
                # prepare new nodes data
                left_node = struct.pack('<' + self.node_heading_format + self.pointer_format +
                                        old_node_size * (self.
                                                         key_format + self.pointer_format),
                                        old_node_size,
                                        children_flag,
                                        *old_node_data[:old_node_size * 2 + 1])
                if nr_of_keys_to_rewrite:
                    keys_before = old_node_data[(old_node_size +
                                                 1) * 2:-nr_of_keys_to_rewrite * 2]
                    keys_after = old_node_data[-nr_of_keys_to_rewrite * 2:]
                else:
                    keys_before = old_node_data[(old_node_size + 1) * 2:]
                    keys_after = []
                right_node = struct.pack('<' + self.node_heading_format + self.pointer_format +
                                         (new_node_size - nr_of_keys_to_rewrite - 1) * (self.
                                                                                        key_format + self.pointer_format),
                                         new_node_size,
                                         children_flag,
                                         *keys_before)
                right_node += struct.pack(
                    '<' + (nr_of_keys_to_rewrite + 1) *
                    (self.key_format + self.pointer_format),
                    new_key,
                    new_pointer,
                    *keys_after)
            new_root = self._prepare_new_root_data(key_moved_to_root,
                                                   new_node_start,
                                                   new_node_start + self.node_size)
            left_node += (self.node_capacity - old_node_size) * \
                (self.key_size + self.pointer_size) * '\x00'
            # adding blanks after new node
            right_node += (self.node_capacity - new_node_size) * \
                (self.key_size + self.pointer_size) * '\x00'
            self.buckets.seek(0, 2)
            self.buckets.write(left_node + right_node)
            self.buckets.seek(self.data_start)
            self.buckets.write(new_root)

            self._read_single_node_key.delete(node_start)
            self._read_node_nr_of_elements_and_children_flag.delete(node_start)
            return None

    def _split_node(self, node_start, nr_of_keys_to_rewrite, new_key, new_pointer, children_flag, create_new_root=False):
        """
        Splits full node in two separate ones, first half of records stays on old position,
        second half is written as new leaf at the end of file.
        """
        half_size = self.node_capacity / 2
        if self.node_capacity % 2 == 0:
            old_node_size = new_node_size = half_size
        else:
            old_node_size = half_size
            new_node_size = half_size + 1
        if create_new_root:
            self._create_new_root_from_node(node_start, children_flag, nr_of_keys_to_rewrite, new_node_size, old_node_size, new_key, new_pointer)
        else:
            blanks = (self.node_capacity - new_node_size) * (
                self.key_size + self.pointer_size) * '\x00'
            if nr_of_keys_to_rewrite == new_node_size:  # insert key into first half of node
                # reading second half of node
                self.buckets.seek(self._calculate_key_position(node_start,
                                                               old_node_size,
                                                               'n') + self.pointer_size)
                # read all keys with key>new_key
                data = self.buckets.read(nr_of_keys_to_rewrite *
                                         (self.key_size + self.pointer_size))
                old_node_data = struct.unpack('<' + nr_of_keys_to_rewrite *
                                              (self.key_format + self.pointer_format), data)
                # write new node at end of file
                self.buckets.seek(0, 2)
                new_node_start = self.buckets.tell()
                # prepare new node_data
                new_node = struct.pack('<' + self.node_heading_format + self.pointer_format +
                                       (self.key_format +
                                        self.pointer_format) * new_node_size,
                                       new_node_size,
                                       children_flag,
                                       new_pointer,
                                       *old_node_data)
                new_node += blanks
                # write new node
                self.buckets.write(new_node)
                # update old node data
                self._update_size(
                    node_start, old_node_size)

                self._read_single_node_key.delete(node_start)
                self._read_node_nr_of_elements_and_children_flag.delete(
                    node_start)

                return new_node_start, new_key
            elif nr_of_keys_to_rewrite > half_size:  # insert key into first half of node
                # seek for first key to rewrite
                self.buckets.seek(self._calculate_key_position(node_start, self.node_capacity - nr_of_keys_to_rewrite, 'n')
                                  + self.pointer_size)
                # read all keys with key>new_key
                data = self.buckets.read(
                    nr_of_keys_to_rewrite * (self.key_size + self.pointer_size))
                old_node_data = struct.unpack(
                    '<' + nr_of_keys_to_rewrite * (self.key_format + self.pointer_format), data)
                key_moved_to_parent_node = old_node_data[-(
                    new_node_size + 1) * 2]
                self.buckets.seek(0, 2)
                new_node_start = self.buckets.tell()
                # prepare new node_data
                new_node = struct.pack('<' + self.node_heading_format +
                                       self.pointer_format + (self.key_format +
                                                              self.pointer_format) * new_node_size,
                                       new_node_size,
                                       children_flag,
                                       old_node_data[-new_node_size * 2 - 1],
                                       *old_node_data[-new_node_size * 2:])
                new_node += blanks
                # write new node
                self.buckets.write(new_node)
                self._update_size(
                    node_start, old_node_size)
                # seek position of new key in first half
                self.buckets.seek(self._calculate_key_position(node_start, self.node_capacity - nr_of_keys_to_rewrite, 'n')
                                  + self.pointer_size)
                # write new key and keys after
                self.buckets.write(
                    struct.pack(
                        '<' + (self.key_format + self.pointer_format) *
                        (nr_of_keys_to_rewrite - new_node_size),
                        new_key,
                        new_pointer,
                        *old_node_data[:-(new_node_size + 1) * 2]))

                self._read_single_node_key.delete(node_start)
                self._read_node_nr_of_elements_and_children_flag.delete(
                    node_start)

                return new_node_start, key_moved_to_parent_node
            else:  # key goes into second half
                # reading second half of node
                self.buckets.seek(self._calculate_key_position(node_start,
                                                               old_node_size,
                                                               'n')
                                  + self.pointer_size)
                data = self.buckets.read(
                    new_node_size * (self.key_size + self.pointer_size))
                old_node_data = struct.unpack('<' + new_node_size *
                                              (self.key_format + self.pointer_format), data)
                # find key which goes to parent node
                key_moved_to_parent_node = old_node_data[0]
                self.buckets.seek(0, 2)  # end of file
                new_node_start = self.buckets.tell()
                index_of_records_split = nr_of_keys_to_rewrite * 2
                # prepare new node_data
                first_leaf_pointer = old_node_data[1]
                old_node_data = old_node_data[2:]
                if index_of_records_split:
                    keys_before = old_node_data[:-index_of_records_split]
                    keys_after = old_node_data[-index_of_records_split:]
                else:
                    keys_before = old_node_data
                    keys_after = []
                new_node = struct.pack('<' + self.node_heading_format + self.pointer_format +
                                       (self.key_format + self.pointer_format) *
                                       (new_node_size -
                                        nr_of_keys_to_rewrite - 1),
                                       new_node_size,
                                       children_flag,
                                       first_leaf_pointer,
                                       *keys_before)
                new_node += struct.pack('<' + (self.key_format + self.pointer_format) *
                                        (nr_of_keys_to_rewrite + 1),
                                        new_key,
                                        new_pointer,
                                        *keys_after)
                new_node += blanks
                # write new node
                self.buckets.write(new_node)
                self._update_size(node_start, old_node_size)

                self._read_single_node_key.delete(node_start)
                self._read_node_nr_of_elements_and_children_flag.delete(
                    node_start)

                return new_node_start, key_moved_to_parent_node

    def insert_first_record_into_leaf(self, leaf_start, key, doc_id, start, size, status):
        self.buckets.seek(leaf_start)
        self.buckets.write(struct.pack('<' + self.elements_counter_format,
                                       1))
        self.buckets.seek(leaf_start + self.leaf_heading_size)
        self.buckets.write(struct.pack('<' + self.single_leaf_record_format,
                                       key,
                                       doc_id,
                                       start,
                                       size,
                                       status))

#            self._read_single_leaf_record.delete(leaf_start)
        self._find_key_in_leaf.delete(leaf_start)
        self._read_leaf_nr_of_elements.delete(leaf_start)
        self._read_leaf_nr_of_elements_and_neighbours.delete(leaf_start)

    def _insert_new_record_into_leaf(self, leaf_start, key, doc_id, start, size, status, nodes_stack, indexes):
        nr_of_elements = self._read_leaf_nr_of_elements(leaf_start)
        if nr_of_elements == 0:
            self.insert_first_record_into_leaf(
                leaf_start, key, doc_id, start, size, status)
            return
        leaf_start, new_record_position, nr_of_records_to_rewrite, full_leaf, on_deleted\
            = self._find_place_in_leaf(key, leaf_start, nr_of_elements)
        if full_leaf:
            try:  # check if leaf has parent node
                leaf_parent_pointer = nodes_stack.pop()
            except IndexError:  # leaf is a root
                leaf_parent_pointer = 0
            split_data = self._split_leaf(leaf_start,
                                          nr_of_records_to_rewrite,
                                          key,
                                          doc_id,
                                          start,
                                          size,
                                          status,
                                          create_new_root=(False if leaf_parent_pointer else True))
            if split_data is not None:  # means that split created new root or replaced split with update_if_has_deleted
                new_leaf_start_position, key_moved_to_parent_node = split_data
                self._insert_new_key_into_node(leaf_parent_pointer,
                                               key_moved_to_parent_node,
                                               leaf_start,
                                               new_leaf_start_position,
                                               nodes_stack,
                                               indexes)
        else:  # there is a place for record in leaf
            self.buckets.seek(leaf_start)
            self._update_leaf(
                leaf_start, new_record_position, nr_of_elements, nr_of_records_to_rewrite,
                on_deleted, key, doc_id, start, size, status)

    def _update_node(self, new_key_position, nr_of_keys_to_rewrite, new_key, new_pointer):
        if nr_of_keys_to_rewrite == 0:
            self.buckets.seek(new_key_position)
            self.buckets.write(
                struct.pack('<' + self.key_format + self.pointer_format,
                            new_key,
                            new_pointer))
            self.flush()
        else:
            self.buckets.seek(new_key_position)
            data = self.buckets.read(nr_of_keys_to_rewrite * (
                                     self.key_size + self.pointer_size))
            keys_to_rewrite = struct.unpack(
                '<' + nr_of_keys_to_rewrite * (self.key_format + self.pointer_format), data)
            self.buckets.seek(new_key_position)
            self.buckets.write(
                struct.pack(
                    '<' + (nr_of_keys_to_rewrite + 1) *
                    (self.key_format + self.pointer_format),
                    new_key,
                    new_pointer,
                    *keys_to_rewrite))
            self.flush()

    def _insert_new_key_into_node(self, node_start, new_key, old_half_start, new_half_start, nodes_stack, indexes):
        parent_key_index = indexes.pop()
        nr_of_elements, children_flag = self._read_node_nr_of_elements_and_children_flag(node_start)
        parent_prev_pointer = self._read_single_node_key(
            node_start, parent_key_index)[0]
        if parent_prev_pointer == old_half_start:  # splited child was on the left side of his parent key, must write new key before it
            new_key_position = self.pointer_size + self._calculate_key_position(node_start, parent_key_index, 'n')
            nr_of_keys_to_rewrite = nr_of_elements - parent_key_index
        else:  # splited child was on the right side of his parent key, must write new key after it
            new_key_position = self.pointer_size + self._calculate_key_position(node_start, parent_key_index + 1, 'n')
            nr_of_keys_to_rewrite = nr_of_elements - (parent_key_index + 1)
        if nr_of_elements == self.node_capacity:
            try:  # check if node has parent
                node_parent_pointer = nodes_stack.pop()
            except IndexError:  # node is a root
                node_parent_pointer = 0
            new_data = self._split_node(node_start,
                                        nr_of_keys_to_rewrite,
                                        new_key,
                                        new_half_start,
                                        children_flag,
                                        create_new_root=(False if node_parent_pointer else True))
            if new_data:  # if not new_data, new root has been created
                new_node_start_position, key_moved_to_parent_node = new_data
                self._insert_new_key_into_node(node_parent_pointer,
                                               key_moved_to_parent_node,
                                               node_start,
                                               new_node_start_position,
                                               nodes_stack,
                                               indexes)

            self._find_first_key_occurence_in_node.delete(node_start)
            self._find_last_key_occurence_in_node.delete(node_start)
        else:  # there is a empty slot for new key in node
            self._update_size(node_start, nr_of_elements + 1)
            self._update_node(new_key_position,
                              nr_of_keys_to_rewrite,
                              new_key,
                              new_half_start)

            self._find_first_key_occurence_in_node.delete(node_start)
            self._find_last_key_occurence_in_node.delete(node_start)
            self._read_single_node_key.delete(node_start)
            self._read_node_nr_of_elements_and_children_flag.delete(node_start)

    def _find_leaf_to_insert(self, key):
        """
        Traverses tree in search for leaf for insert, remembering parent nodes in path,
        looks for last occurence of key if already in tree.
        """
        nodes_stack = [self.data_start]
        if self.root_flag == 'l':
            return nodes_stack, []
        else:
            nr_of_elements, curr_child_flag = self._read_node_nr_of_elements_and_children_flag(self.data_start)
            curr_index, curr_pointer = self._find_last_key_occurence_in_node(
                self.data_start, key, nr_of_elements)
            nodes_stack.append(curr_pointer)
            indexes = [curr_index]
            while(curr_child_flag == 'n'):
                nr_of_elements, curr_child_flag = self._read_node_nr_of_elements_and_children_flag(curr_pointer)
                curr_index, curr_pointer = self._find_last_key_occurence_in_node(curr_pointer, key, nr_of_elements)
                nodes_stack.append(curr_pointer)
                indexes.append(curr_index)
            return nodes_stack, indexes
        # nodes stack contains start addreses of nodes directly above leaf with key, indexes match keys adjacent nodes_stack values (as pointers)
        # required when inserting new keys in upper tree levels

    def _find_leaf_with_last_key_occurence(self, key):
        if self.root_flag == 'l':
            return self.data_start
        else:
            nr_of_elements, curr_child_flag = self._read_node_nr_of_elements_and_children_flag(self.data_start)
            curr_position = self._find_last_key_occurence_in_node(
                self.data_start, key, nr_of_elements)[1]
            while(curr_child_flag == 'n'):
                nr_of_elements, curr_child_flag = self._read_node_nr_of_elements_and_children_flag(curr_position)
                curr_position = self._find_last_key_occurence_in_node(
                    curr_position, key, nr_of_elements)[1]
            return curr_position

    def _find_leaf_with_first_key_occurence(self, key):
        if self.root_flag == 'l':
            return self.data_start
        else:
            nr_of_elements, curr_child_flag = self._read_node_nr_of_elements_and_children_flag(self.data_start)
            curr_position = self._find_first_key_occurence_in_node(
                self.data_start, key, nr_of_elements)[1]
            while(curr_child_flag == 'n'):
                nr_of_elements, curr_child_flag = self._read_node_nr_of_elements_and_children_flag(curr_position)
                curr_position = self._find_first_key_occurence_in_node(
                    curr_position, key, nr_of_elements)[1]
            return curr_position

    def _find_key(self, key):
        containing_leaf_start = self._find_leaf_with_first_key_occurence(key)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(containing_leaf_start)
        try:
            doc_id, l_key, start, size, status = self._find_key_in_leaf(
                containing_leaf_start, key, nr_of_elements)
        except ElemNotFound:
            if next_leaf:
                nr_of_elements = self._read_leaf_nr_of_elements(next_leaf)
            else:
                raise ElemNotFound
            doc_id, l_key, start, size, status = self._find_key_in_leaf(
                next_leaf, key, nr_of_elements)
        return doc_id, l_key, start, size, status

    def _find_key_to_update(self, key, doc_id):
        """
        Search tree for key that matches not only given key but also doc_id.
        """
        containing_leaf_start = self._find_leaf_with_first_key_occurence(key)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(containing_leaf_start)
        try:
            leaf_start, record_index, doc_id, l_key, start, size, status = self._find_key_in_leaf_for_update(key,
                                                                                                             doc_id,
                                                                                                             containing_leaf_start,
                                                                                                             nr_of_elements)
        except ElemNotFound:
            if next_leaf:
                nr_of_elements = self._read_leaf_nr_of_elements(next_leaf)
            else:
                raise TryReindexException()
            try:
                leaf_start, record_index, doc_id, l_key, start, size, status = self._find_key_in_leaf_for_update(key,
                                                                                                                 doc_id,
                                                                                                                 next_leaf,
                                                                                                                 nr_of_elements)
            except ElemNotFound:
                raise TryReindexException()
        return leaf_start, record_index, doc_id, l_key, start, size, status

    def update(self, doc_id, key, u_start=0, u_size=0, u_status='o'):
        containing_leaf_start, element_index, old_doc_id, old_key, old_start, old_size, old_status = self._find_key_to_update(key, doc_id)
        if u_start:
            old_start = u_start
        if u_size:
            old_size = u_size
        if u_status:
            old_status = u_status
        new_data = (old_doc_id, old_start, old_size, old_status)
        self._update_element(containing_leaf_start, element_index, new_data)

        self._find_key.delete(key)
        self._match_doc_id.delete(doc_id)
        self._find_key_in_leaf.delete(containing_leaf_start, key)
        return True

    def delete(self, doc_id, key, start=0, size=0):
        containing_leaf_start, element_index = self._find_key_to_update(
            key, doc_id)[:2]
        self._delete_element(containing_leaf_start, element_index)

        self._find_key.delete(key)
        self._match_doc_id.delete(doc_id)
        self._find_key_in_leaf.delete(containing_leaf_start, key)
        return True

    def _find_key_many(self, key, limit=1, offset=0):
        leaf_with_key = self._find_leaf_with_first_key_occurence(key)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        try:
            leaf_with_key, key_index = self._find_index_of_first_key_equal(
                key, leaf_with_key, nr_of_elements)
            nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        except ElemNotFound:
            leaf_with_key = next_leaf
            key_index = 0
            nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        while offset:
            if key_index < nr_of_elements:
                curr_key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if key == curr_key:
                    if status != 'd':
                        offset -= 1
                    key_index += 1
                else:
                    return
            else:
                key_index = 0
                if next_leaf:
                    leaf_with_key = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return
        while limit:
            if key_index < nr_of_elements:
                curr_key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if key == curr_key:
                    if status != 'd':
                        yield doc_id, start, size, status
                        limit -= 1
                    key_index += 1
                else:
                    return
            else:
                key_index = 0
                if next_leaf:
                    leaf_with_key = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return

    def _find_key_smaller(self, key, limit=1, offset=0):
        leaf_with_key = self._find_leaf_with_first_key_occurence(key)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        leaf_with_key, key_index = self._find_index_of_first_key_equal_or_smaller_key(key, leaf_with_key, nr_of_elements)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        curr_key = self._read_single_leaf_record(leaf_with_key, key_index)[0]
        if curr_key >= key:
            key_index -= 1
        while offset:
            if key_index >= 0:
                key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if status != 'd':
                    offset -= 1
                key_index -= 1
            else:
                if prev_leaf:
                    leaf_with_key = prev_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(prev_leaf)
                    key_index = nr_of_elements - 1
                else:
                    return
        while limit:
            if key_index >= 0:
                key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if status != 'd':
                    yield doc_id, key, start, size, status
                    limit -= 1
                key_index -= 1
            else:
                if prev_leaf:
                    leaf_with_key = prev_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(prev_leaf)
                    key_index = nr_of_elements - 1
                else:
                    return

    def _find_key_equal_and_smaller(self, key, limit=1, offset=0):
        leaf_with_key = self._find_leaf_with_last_key_occurence(key)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        try:
            leaf_with_key, key_index = self._find_index_of_last_key_equal_or_smaller_key(key, leaf_with_key, nr_of_elements)
            nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        except ElemNotFound:
            leaf_with_key = prev_leaf
            key_index = self._read_leaf_nr_of_elements_and_neighbours(
                leaf_with_key)[0]
        curr_key = self._read_single_leaf_record(leaf_with_key, key_index)[0]
        if curr_key > key:
            key_index -= 1
        while offset:
            if key_index >= 0:
                key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if status != 'd':
                    offset -= 1
                key_index -= 1
            else:
                if prev_leaf:
                    leaf_with_key = prev_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(prev_leaf)
                    key_index = nr_of_elements - 1
                else:
                    return
        while limit:
            if key_index >= 0:
                key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if status != 'd':
                    yield doc_id, key, start, size, status
                    limit -= 1
                key_index -= 1
            else:
                if prev_leaf:
                    leaf_with_key = prev_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(prev_leaf)
                    key_index = nr_of_elements - 1
                else:
                    return

    def _find_key_bigger(self, key, limit=1, offset=0):
        leaf_with_key = self._find_leaf_with_last_key_occurence(key)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        try:
            leaf_with_key, key_index = self._find_index_of_last_key_equal_or_smaller_key(key, leaf_with_key, nr_of_elements)
            nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        except ElemNotFound:
            key_index = 0
        curr_key = self._read_single_leaf_record(leaf_with_key, key_index)[0]
        if curr_key <= key:
            key_index += 1
        while offset:
            if key_index < nr_of_elements:
                curr_key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if status != 'd':
                    offset -= 1
                key_index += 1
            else:
                key_index = 0
                if next_leaf:
                    leaf_with_key = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return
        while limit:
            if key_index < nr_of_elements:
                curr_key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if status != 'd':
                    yield doc_id, curr_key, start, size, status
                    limit -= 1
                key_index += 1
            else:
                key_index = 0
                if next_leaf:
                    leaf_with_key = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return

    def _find_key_equal_and_bigger(self, key, limit=1, offset=0):
        leaf_with_key = self._find_leaf_with_first_key_occurence(key)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        leaf_with_key, key_index = self._find_index_of_first_key_equal_or_smaller_key(key, leaf_with_key, nr_of_elements)
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
        curr_key = self._read_single_leaf_record(leaf_with_key, key_index)[0]
        if curr_key < key:
            key_index += 1
        while offset:
            if key_index < nr_of_elements:
                curr_key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if status != 'd':
                    offset -= 1
                key_index += 1
            else:
                key_index = 0
                if next_leaf:
                    leaf_with_key = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return
        while limit:
            if key_index < nr_of_elements:
                curr_key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_with_key, key_index)
                if status != 'd':
                    yield doc_id, curr_key, start, size, status
                    limit -= 1
                key_index += 1
            else:
                key_index = 0
                if next_leaf:
                    leaf_with_key = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return

    def _find_key_between(self, start, end, limit, offset, inclusive_start, inclusive_end):
        """
        Returns generator containing all keys withing given interval.
        """
        if inclusive_start:
            leaf_with_key = self._find_leaf_with_first_key_occurence(start)
            nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
            leaf_with_key, key_index = self._find_index_of_first_key_equal_or_smaller_key(start, leaf_with_key, nr_of_elements)
            nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
            curr_key = self._read_single_leaf_record(
                leaf_with_key, key_index)[0]
            if curr_key < start:
                key_index += 1
        else:
            leaf_with_key = self._find_leaf_with_last_key_occurence(start)
            nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_with_key)
            leaf_with_key, key_index = self._find_index_of_last_key_equal_or_smaller_key(start, leaf_with_key, nr_of_elements)
            curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_with_key, key_index)
            if curr_key <= start:
                key_index += 1
        while offset:
            if key_index < nr_of_elements:
                curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_with_key, key_index)
                if curr_status != 'd':
                    offset -= 1
                key_index += 1
            else:
                key_index = 0
                if next_leaf:
                    leaf_with_key = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return
        while limit:
            if key_index < nr_of_elements:
                curr_key, curr_doc_id, curr_start, curr_size, curr_status = self._read_single_leaf_record(leaf_with_key, key_index)
                if curr_key > end or (curr_key == end and not inclusive_end):
                    return
                elif curr_status != 'd':
                    yield curr_doc_id, curr_key, curr_start, curr_size, curr_status
                    limit -= 1
                key_index += 1
            else:
                key_index = 0
                if next_leaf:
                    leaf_with_key = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return

    def get(self, key):
        return self._find_key(self.make_key(key))

    def get_many(self, key, limit=1, offset=0):
        return self._find_key_many(self.make_key(key), limit, offset)

    def get_between(self, start, end, limit=1, offset=0, inclusive_start=True, inclusive_end=True):
        if start is None:
            end = self.make_key(end)
            if inclusive_end:
                return self._find_key_equal_and_smaller(end, limit, offset)
            else:
                return self._find_key_smaller(end, limit, offset)
        elif end is None:
            start = self.make_key(start)
            if inclusive_start:
                return self._find_key_equal_and_bigger(start, limit, offset)
            else:
                return self._find_key_bigger(start, limit, offset)
        else:
            start = self.make_key(start)
            end = self.make_key(end)
            return self._find_key_between(start, end, limit, offset, inclusive_start, inclusive_end)

    def all(self, limit=-1, offset=0):
        """
        Traverses linked list of all tree leaves and returns generator containing all elements stored in index.
        """
        if self.root_flag == 'n':
            leaf_start = self.data_start + self.node_size
        else:
            leaf_start = self.data_start
        nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(leaf_start)
        key_index = 0
        while offset:
            if key_index < nr_of_elements:
                curr_key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_start, key_index)
                if status != 'd':
                    offset -= 1
                key_index += 1
            else:
                key_index = 0
                if next_leaf:
                    leaf_start = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return
        while limit:
            if key_index < nr_of_elements:
                curr_key, doc_id, start, size, status = self._read_single_leaf_record(
                    leaf_start, key_index)
                if status != 'd':
                    yield doc_id, curr_key, start, size, status
                    limit -= 1
                key_index += 1
            else:
                key_index = 0
                if next_leaf:
                    leaf_start = next_leaf
                    nr_of_elements, prev_leaf, next_leaf = self._read_leaf_nr_of_elements_and_neighbours(next_leaf)
                else:
                    return

    def make_key(self, key):
        raise NotImplementedError()

    def make_key_value(self, data):
        raise NotImplementedError()

    def _open_storage(self):
        s = globals()[self.storage_class]
        if not self.storage:
            self.storage = s(self.db_path, self.name)
        self.storage.open()

    def _create_storage(self):
        s = globals()[self.storage_class]
        if not self.storage:
            self.storage = s(self.db_path, self.name)
        self.storage.create()

    def compact(self, node_capacity=0):
        if not node_capacity:
            node_capacity = self.node_capacity

        compact_ind = self.__class__(
            self.db_path, self.name + '_compact', node_capacity=node_capacity)
        compact_ind.create_index()

        gen = self.all()
        while True:
            try:
                doc_id, key, start, size, status = next(gen)
            except StopIteration:
                break
            self.storage._f.seek(start)
            value = self.storage._f.read(size)
            start_ = compact_ind.storage._f.tell()
            compact_ind.storage._f.write(value)
            compact_ind.insert(doc_id, key, start_, size, status)

        compact_ind.close_index()
        original_name = self.name
        # os.unlink(os.path.join(self.db_path, self.name + "_buck"))
        self.close_index()
        shutil.move(os.path.join(compact_ind.db_path, compact_ind.
                                 name + "_buck"), os.path.join(self.db_path, self.name + "_buck"))
        shutil.move(os.path.join(compact_ind.db_path, compact_ind.
                                 name + "_stor"), os.path.join(self.db_path, self.name + "_stor"))
        # self.name = original_name
        self.open_index()  # reload...
        self.name = original_name
        self._save_params(dict(name=original_name))
        self._fix_params()
        self._clear_cache()
        return True

    def _fix_params(self):
        super(IU_TreeBasedIndex, self)._fix_params()
        self._count_props()

    def _clear_cache(self):
        self._find_key.clear()
        self._match_doc_id.clear()
#        self._read_single_leaf_record.clear()
        self._find_key_in_leaf.clear()
        self._read_single_node_key.clear()
        self._find_first_key_occurence_in_node.clear()
        self._find_last_key_occurence_in_node.clear()
        self._read_leaf_nr_of_elements.clear()
        self._read_leaf_neighbours.clear()
        self._read_leaf_nr_of_elements_and_neighbours.clear()
        self._read_node_nr_of_elements_and_children_flag.clear()

    def close_index(self):
        super(IU_TreeBasedIndex, self).close_index()
        self._clear_cache()


class IU_MultiTreeBasedIndex(IU_TreeBasedIndex):

    """
    Class that allows to index more than one key per database record.

    It operates very well on GET/INSERT. It's not optimized for
    UPDATE operations (will always readd everything)
    """

    def __init__(self, *args, **kwargs):
        super(IU_MultiTreeBasedIndex, self).__init__(*args, **kwargs)

    def insert(self, doc_id, key, start, size, status='o'):
        if isinstance(key, (list, tuple)):
            key = set(key)
        elif not isinstance(key, set):
            key = set([key])
        ins = super(IU_MultiTreeBasedIndex, self).insert
        for curr_key in key:
            ins(doc_id, curr_key, start, size, status)
        return True

    def update(self, doc_id, key, u_start, u_size, u_status='o'):
        if isinstance(key, (list, tuple)):
            key = set(key)
        elif not isinstance(key, set):
            key = set([key])
        upd = super(IU_MultiTreeBasedIndex, self).update
        for curr_key in key:
            upd(doc_id, curr_key, u_start, u_size, u_status)

    def delete(self, doc_id, key, start=0, size=0):
        if isinstance(key, (list, tuple)):
            key = set(key)
        elif not isinstance(key, set):
            key = set([key])
        delete = super(IU_MultiTreeBasedIndex, self).delete
        for curr_key in key:
            delete(doc_id, curr_key, start, size)

    def get(self, key):
        return super(IU_MultiTreeBasedIndex, self).get(key)

    def make_key_value(self, data):
        raise NotImplementedError()


# classes for public use, done in this way because of
# generation static files with indexes (_index directory)


class TreeBasedIndex(IU_TreeBasedIndex):
    pass


class MultiTreeBasedIndex(IU_MultiTreeBasedIndex):

    """
    It allows to index more than one key for record. (ie. prefix/infix/suffix search mechanizms)
    That class is designed to be used in custom indexes.
    """
    pass
